@@ -17,9 +17,9 @@ MVMObject * MVM_repr_alloc_init(MVMThreadContext *tc, MVMObject *type) {
     MVMObject *obj = REPR(type)->allocate(tc, STABLE(type));
 
     if (REPR(obj)->initialize) {
-        MVMROOT(tc, obj, {
+        MVMROOT(tc, obj) {
             REPR(obj)->initialize(tc, STABLE(obj), obj, OBJECT_BODY(obj));
-        });
+        }
     }
 
     return obj;
@@ -66,16 +66,16 @@ void MVM_repr_set_dimensions(MVMThreadContext *tc, MVMObject *obj, MVMObject *di
             OBJECT_BODY(obj), num_dims, tc->multi_dim_indices);
     }
     else {
-        MVM_exception_throw_adhoc(tc, "Cannot set dimensions on a type object");
+        MVM_exception_throw_adhoc(tc, "Cannot set dimensions on a type object. Did you forget a '.new'?");
     }
 }
 
 MVM_PUBLIC MVMObject * MVM_repr_pos_slice(MVMThreadContext *tc, MVMObject *src, MVMint64 start, MVMint64 end) {
     MVMObject *dest = NULL;
-    MVMROOT(tc, src, {
+    MVMROOT(tc, src) {
         dest = MVM_repr_alloc_init(tc, src);
         REPR(src)->pos_funcs.slice(tc, STABLE(src), src, OBJECT_BODY(src), dest, start, end);
-    });
+    }
     return dest;
 }
 
@@ -95,8 +95,12 @@ MVM_PUBLIC MVMint64 MVM_repr_exists_pos(MVMThreadContext *tc, MVMObject *obj, MV
 MVMint64 MVM_repr_at_pos_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
     MVMRegister value;
     if (REPR(obj)->ID == MVM_REPR_ID_VMArray) {
-        MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
-            idx, &value, MVM_reg_int64);
+        if (((MVMArrayREPRData *)STABLE(obj)->REPR_data)->slot_type == MVM_ARRAY_I64) {
+            MVM_VMArray_at_pos_i(tc, STABLE(obj), obj, OBJECT_BODY(obj), idx, &value);
+        }
+        else {
+            MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj), idx, &value, MVM_reg_int64);
+        }
     }
     else {
         REPR(obj)->pos_funcs.at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
@@ -105,11 +109,32 @@ MVMint64 MVM_repr_at_pos_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
     return value.i64;
 }
 
+MVMuint64 MVM_repr_at_pos_u(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
+    MVMRegister value;
+    if (REPR(obj)->ID == MVM_REPR_ID_VMArray) {
+        if (((MVMArrayREPRData *)STABLE(obj)->REPR_data)->slot_type == MVM_ARRAY_U64) {
+            MVM_VMArray_at_pos_u(tc, STABLE(obj), obj, OBJECT_BODY(obj), idx, &value);
+        }
+        else {
+            MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj), idx, &value, MVM_reg_uint64);
+        }
+    }
+    else {
+        REPR(obj)->pos_funcs.at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+            idx, &value, MVM_reg_uint64);
+    }
+    return value.u64;
+}
+
 MVMnum64 MVM_repr_at_pos_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
     MVMRegister value;
     if (REPR(obj)->ID == MVM_REPR_ID_VMArray) {
-        MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
-            idx, &value, MVM_reg_num64);
+        if (((MVMArrayREPRData *)STABLE(obj)->REPR_data)->slot_type == MVM_ARRAY_N64) {
+            MVM_VMArray_at_pos_n(tc, STABLE(obj), obj, OBJECT_BODY(obj), idx, &value);
+        }
+        else {
+            MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj), idx, &value, MVM_reg_num64);
+        }
     }
     else {
         REPR(obj)->pos_funcs.at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
@@ -121,8 +146,8 @@ MVMnum64 MVM_repr_at_pos_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
 MVMString * MVM_repr_at_pos_s(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx) {
     MVMRegister value;
     if (REPR(obj)->ID == MVM_REPR_ID_VMArray) {
-        MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
-            idx, &value, MVM_reg_str);
+        MVM_VMArray_at_pos_s(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+            idx, &value);
     }
     else {
         REPR(obj)->pos_funcs.at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
@@ -135,8 +160,8 @@ MVMObject * MVM_repr_at_pos_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx
     if (MVM_LIKELY(IS_CONCRETE(obj))) {
         MVMRegister value;
         if (REPR(obj)->ID == MVM_REPR_ID_VMArray) {
-            MVM_VMArray_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
-                idx, &value, MVM_reg_obj);
+            MVM_VMArray_at_pos_o(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+                idx, &value);
         }
         else if (REPR(obj)->ID == MVM_REPR_ID_P6opaque) {
             MVM_P6opaque_at_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
@@ -164,6 +189,12 @@ MVMint64 MVM_repr_at_pos_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObj
     return r.i64;
 }
 
+MVMuint64 MVM_repr_at_pos_multidim_u(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
+    MVMRegister r;
+    at_pos_multidim(tc, obj, indices, &r, MVM_reg_uint64);
+    return r.u64;
+}
+
 MVMnum64 MVM_repr_at_pos_multidim_n(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices) {
     MVMRegister r;
     at_pos_multidim(tc, obj, indices, &r, MVM_reg_num64);
@@ -194,6 +225,12 @@ MVMint64 MVM_repr_at_pos_2d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx
     return r.i64;
 }
 
+MVMuint64 MVM_repr_at_pos_2d_u(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2) {
+    MVMRegister r;
+    at_pos_2d(tc, obj, idx1, idx2, &r, MVM_reg_uint64);
+    return r.u64;
+}
+
 MVMnum64 MVM_repr_at_pos_2d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2) {
     MVMRegister r;
     at_pos_2d(tc, obj, idx1, idx2, &r, MVM_reg_num64);
@@ -224,6 +261,12 @@ MVMint64 MVM_repr_at_pos_3d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx
     return r.i64;
 }
 
+MVMuint64 MVM_repr_at_pos_3d_u(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3) {
+    MVMRegister r;
+    at_pos_3d(tc, obj, idx1, idx2, idx3, &r, MVM_reg_uint64);
+    return r.u64;
+}
+
 MVMnum64 MVM_repr_at_pos_3d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3) {
     MVMRegister r;
     at_pos_3d(tc, obj, idx1, idx2, idx3, &r, MVM_reg_num64);
@@ -249,6 +292,13 @@ void MVM_repr_bind_pos_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVM
         idx, val, MVM_reg_int64);
 }
 
+void MVM_repr_bind_pos_u(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVMuint64 value) {
+    MVMRegister val;
+    val.u64 = value;
+    REPR(obj)->pos_funcs.bind_pos(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+        idx, val, MVM_reg_uint64);
+}
+
 void MVM_repr_bind_pos_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVMnum64 value) {
     MVMRegister val;
     val.n64 = value;
@@ -270,6 +320,14 @@ void MVM_repr_bind_pos_o(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx, MVM
         idx, val, MVM_reg_obj);
 }
 
+void MVM_repr_write_buf(MVMThreadContext *tc, MVMObject *obj, char *value, MVMint64 offset, MVMint64 size) {
+    REPR(obj)->pos_funcs.write_buf(tc, STABLE(obj), obj, OBJECT_BODY(obj), value, offset, size);
+}
+
+MVMint64 MVM_repr_read_buf(MVMThreadContext *tc, MVMObject *obj, MVMint64 offset, MVMint64 size) {
+    return REPR(obj)->pos_funcs.read_buf(tc, STABLE(obj), obj, OBJECT_BODY(obj), offset, size);
+}
+
 static void bind_pos_multidim(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMRegister value, MVMuint16 kind) {
     MVMint64 num_indices;
     MVM_repr_populate_indices_array(tc, indices, &num_indices);
@@ -282,6 +340,11 @@ void MVM_repr_bind_pos_multidim_i(MVMThreadContext *tc, MVMObject *obj, MVMObjec
     r.i64 = value;
     bind_pos_multidim(tc, obj, indices, r, MVM_reg_int64);
 }
+void MVM_repr_bind_pos_multidim_u(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMuint64 value) {
+    MVMRegister r;
+    r.u64 = value;
+    bind_pos_multidim(tc, obj, indices, r, MVM_reg_uint64);
+}
 void MVM_repr_bind_pos_multidim_n(MVMThreadContext *tc, MVMObject *obj, MVMObject *indices, MVMnum64 value) {
     MVMRegister r;
     r.n64 = value;
@@ -309,6 +372,11 @@ void MVM_repr_bind_pos_2d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1,
     r.i64 = value;
     bind_pos_2d(tc, obj, idx1, idx2, r, MVM_reg_int64);
 }
+void MVM_repr_bind_pos_2d_u(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMuint64 value) {
+    MVMRegister r;
+    r.u64 = value;
+    bind_pos_2d(tc, obj, idx1, idx2, r, MVM_reg_uint64);
+}
 void MVM_repr_bind_pos_2d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMnum64 value) {
     MVMRegister r;
     r.n64 = value;
@@ -336,6 +404,11 @@ void MVM_repr_bind_pos_3d_i(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1,
     r.i64 = value;
     bind_pos_3d(tc, obj, idx1, idx2, idx3, r, MVM_reg_int64);
 }
+void MVM_repr_bind_pos_3d_u(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMuint64 value) {
+    MVMRegister r;
+    r.u64 = value;
+    bind_pos_3d(tc, obj, idx1, idx2, idx3, r, MVM_reg_uint64);
+}
 void MVM_repr_bind_pos_3d_n(MVMThreadContext *tc, MVMObject *obj, MVMint64 idx1, MVMint64 idx2, MVMint64 idx3, MVMnum64 value) {
     MVMRegister r;
     r.n64 = value;
@@ -477,6 +550,19 @@ MVMint64 MVM_repr_at_key_i(MVMThreadContext *tc, MVMObject *obj, MVMString *key)
     return value.i64;
 }
 
+MVMuint64 MVM_repr_at_key_u(MVMThreadContext *tc, MVMObject *obj, MVMString *key) {
+    MVMRegister value;
+    if (REPR(obj)->ID == MVM_REPR_ID_MVMHash) {
+        MVMHash_at_key(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+            (MVMObject *)key, &value, MVM_reg_uint64);
+    }
+    else {
+        REPR(obj)->ass_funcs.at_key(tc, STABLE(obj), obj, OBJECT_BODY(obj),
+                                (MVMObject *)key, &value, MVM_reg_uint64);
+    }
+    return value.u64;
+}
+
 MVMnum64 MVM_repr_at_key_n(MVMThreadContext *tc, MVMObject *obj, MVMString *key) {
     MVMRegister value;
     if (REPR(obj)->ID == MVM_REPR_ID_MVMHash) {
@@ -577,15 +663,16 @@ MVMObject * MVM_repr_dimensions(MVMThreadContext *tc, MVMObject *obj) {
     if (IS_CONCRETE(obj)) {
         MVMint64 num_dims, i;
         MVMint64 *dims;
-        MVMObject *result = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIntArray);
+        MVMObject *result;
         REPR(obj)->pos_funcs.dimensions(tc, STABLE(obj), obj, OBJECT_BODY(obj),
             &num_dims, &dims);
+        result = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIntArray);
         for (i = 0; i < num_dims; i++)
             MVM_repr_bind_pos_i(tc, result, i, dims[i]);
         return result;
     }
     else {
-        MVM_exception_throw_adhoc(tc, "Cannot get dimensions of a type object");
+        MVM_exception_throw_adhoc(tc, "Cannot get dimensions of a type object. Did you forget a '.new'?");
     }
 }
 
@@ -598,7 +685,7 @@ MVMint64 MVM_repr_num_dimensions(MVMThreadContext *tc, MVMObject *obj) {
         return num_dims;
     }
     else {
-        MVM_exception_throw_adhoc(tc, "Cannot get number of dimensions of a type object");
+        MVM_exception_throw_adhoc(tc, "Cannot get number of dimensions of a type object. Did you forget a '.new'?");
     }
 }
 
@@ -660,10 +747,10 @@ MVMObject * MVM_repr_box_num(MVMThreadContext *tc, MVMObject *type, MVMnum64 val
 
 MVMObject * MVM_repr_box_str(MVMThreadContext *tc, MVMObject *type, MVMString *val) {
     MVMObject *res;
-    MVMROOT(tc, val, {
+    MVMROOT(tc, val) {
         res = MVM_repr_alloc_init(tc, type);
         MVM_repr_set_str(tc, res, val);
-    });
+    }
     return res;
 }
 
@@ -677,7 +764,7 @@ MVM_PUBLIC MVMint64 MVM_repr_get_attr_i(MVMThreadContext *tc, MVMObject *object,
                                            MVMString *name, MVMint16 hint) {
     MVMRegister result_reg;
     if (!IS_CONCRETE(object))
-        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object", MVM_6model_get_debug_name(tc, object));
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, object));
     REPR(object)->attr_funcs.get_attribute(tc,
             STABLE(object), object, OBJECT_BODY(object),
             type, name,
@@ -685,11 +772,23 @@ MVM_PUBLIC MVMint64 MVM_repr_get_attr_i(MVMThreadContext *tc, MVMObject *object,
     return result_reg.i64;
 }
 
+MVM_PUBLIC MVMuint64 MVM_repr_get_attr_u(MVMThreadContext *tc, MVMObject *object, MVMObject *type,
+                                           MVMString *name, MVMint16 hint) {
+    MVMRegister result_reg;
+    if (!IS_CONCRETE(object))
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, object));
+    REPR(object)->attr_funcs.get_attribute(tc,
+            STABLE(object), object, OBJECT_BODY(object),
+            type, name,
+            hint, &result_reg, MVM_reg_uint64);
+    return result_reg.u64;
+}
+
 MVM_PUBLIC MVMnum64 MVM_repr_get_attr_n(MVMThreadContext *tc, MVMObject *object, MVMObject *type,
                                            MVMString *name, MVMint16 hint) {
     MVMRegister result_reg;
     if (!IS_CONCRETE(object))
-        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object", MVM_6model_get_debug_name(tc, object));
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, object));
     REPR(object)->attr_funcs.get_attribute(tc,
             STABLE(object), object, OBJECT_BODY(object),
             type, name,
@@ -701,7 +800,7 @@ MVM_PUBLIC MVMString * MVM_repr_get_attr_s(MVMThreadContext *tc, MVMObject *obje
                                            MVMString *name, MVMint16 hint) {
     MVMRegister result_reg;
     if (!IS_CONCRETE(object))
-        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object", MVM_6model_get_debug_name(tc, object));
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, object));
     REPR(object)->attr_funcs.get_attribute(tc,
             STABLE(object), object, OBJECT_BODY(object),
             type, name,
@@ -713,7 +812,7 @@ MVM_PUBLIC MVMObject * MVM_repr_get_attr_o(MVMThreadContext *tc, MVMObject *obje
                                            MVMString *name, MVMint16 hint) {
     MVMRegister result_reg;
     if (!IS_CONCRETE(object))
-        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object", MVM_6model_get_debug_name(tc, object));
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, object));
     REPR(object)->attr_funcs.get_attribute(tc,
             STABLE(object), object, OBJECT_BODY(object),
             type, name,
@@ -725,7 +824,7 @@ MVM_PUBLIC MVMObject * MVM_repr_get_attr_o(MVMThreadContext *tc, MVMObject *obje
 MVM_PUBLIC void MVM_repr_bind_attr_inso(MVMThreadContext *tc, MVMObject *object, MVMObject *type,
                                            MVMString *name, MVMint16 hint, MVMRegister value_reg, MVMuint16 kind) {
     if (!IS_CONCRETE(object))
-        MVM_exception_throw_adhoc(tc, "Cannot bind attributes in a %s type object", MVM_6model_get_debug_name(tc, object));
+        MVM_exception_throw_adhoc(tc, "Cannot bind attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, object));
     REPR(object)->attr_funcs.bind_attribute(tc,
             STABLE(object), object, OBJECT_BODY(object),
             type, name,
@@ -736,7 +835,7 @@ MVM_PUBLIC void MVM_repr_bind_attr_inso(MVMThreadContext *tc, MVMObject *object,
 MVM_PUBLIC MVMint64 MVM_repr_attribute_inited(MVMThreadContext *tc, MVMObject *obj, MVMObject *type,
                                               MVMString *name) {
     if (!IS_CONCRETE(obj))
-        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object", MVM_6model_get_debug_name(tc, obj));
+        MVM_exception_throw_adhoc(tc, "Cannot look up attributes in a %s type object. Did you forget a '.new'?", MVM_6model_get_debug_name(tc, obj));
     return REPR(obj)->attr_funcs.is_attribute_initialized(tc,
         STABLE(obj), OBJECT_BODY(obj),
         type, name, MVM_NO_HINT);
@@ -757,6 +856,7 @@ MVM_PUBLIC void MVM_repr_atomic_bind_attr_o(MVMThreadContext *tc, MVMObject *obj
         OBJECT_BODY(object), type, name, MVM_reg_obj);
     MVM_store(target, value);
     MVM_gc_write_barrier(tc, (MVMCollectable *)object, (MVMCollectable *)value);
+    MVM_SC_WB_OBJ(tc, object);
 }
 
 MVM_PUBLIC MVMObject * MVM_repr_casattr_o(MVMThreadContext *tc, MVMObject *object,
@@ -766,5 +866,6 @@ MVM_PUBLIC MVMObject * MVM_repr_casattr_o(MVMThreadContext *tc, MVMObject *objec
         OBJECT_BODY(object), type, name, MVM_reg_obj);
     MVMObject *witness = (MVMObject *)MVM_casptr(target, expected, value);
     MVM_gc_write_barrier(tc, (MVMCollectable *)object, (MVMCollectable *)value);
+    MVM_SC_WB_OBJ(tc, object);
     return witness;
 }
