@@ -8,11 +8,11 @@ static const MVMREPROps MVMContext_this_repr;
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
     MVMSTable *st = MVM_gc_allocate_stable(tc, &MVMContext_this_repr, HOW);
 
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
         st->size = sizeof(MVMContext);
-    });
+    }
 
     return st->WHAT;
 }
@@ -72,14 +72,17 @@ static void at_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
     MVMSpeshFrameWalker fw;
     MVMRegister *found;
     MVMuint16 found_kind;
+    MVM_gc_root_temp_push(tc, (MVMCollectable**)&name);
     if (!setup_frame_walker(tc, &fw, body) || !MVM_spesh_frame_walker_get_lex(tc, &fw,
             name, &found, &found_kind, 1, NULL)) {
+        MVM_gc_root_temp_pop(tc);
         char *c_name = MVM_string_utf8_encode_C_string(tc, name);
         char *waste[] = { c_name, NULL };
         MVM_exception_throw_adhoc_free(tc, waste,
             "Lexical with name '%s' does not exist in this frame",
             c_name);
     }
+    MVM_gc_root_temp_pop(tc);
     MVM_spesh_frame_walker_cleanup(tc, &fw);
 
     if (found_kind != kind) {
@@ -272,17 +275,18 @@ static void snapshot_frame_callees(MVMThreadContext *tc, MVMFrame *f) {
         MVMSpeshCandidate *cand = f->caller->spesh_cand;
         MVMFrameExtra *extra = MVM_frame_extra(tc, f);
         extra->caller_info_needed = 1;
-        if (cand && cand->num_inlines) {
-            if (cand->jitcode) {
+        extra->caller_pos_needed = 1;
+        if (cand && cand->body.num_inlines) {
+            if (cand->body.jitcode) {
                 if (extra->caller_jit_position)
                     return;
-                extra->caller_jit_position = MVM_jit_code_get_current_position(tc, cand->jitcode, f->caller);
+                extra->caller_jit_position = MVM_jit_code_get_current_position(tc, cand->body.jitcode, f->caller);
             }
             else {
                 if (extra->caller_deopt_idx)
                     return;
                 /* Store with + 1 to avoid semi-predicate problem. */
-                extra->caller_deopt_idx = MVM_spesh_deopt_find_inactive_frame_deopt_idx(tc, f->caller) + 1;
+                extra->caller_deopt_idx = 1 + MVM_spesh_deopt_find_inactive_frame_deopt_idx(tc, f->caller, cand);
             }
         }
         f = f->caller;
@@ -294,10 +298,23 @@ MVMObject * MVM_context_from_frame(MVMThreadContext *tc, MVMFrame *f) {
     MVMObject *ctx;
     f = MVM_frame_force_to_heap(tc, f);
     snapshot_frame_callees(tc, f);
-    MVMROOT(tc, f, {
+    MVMROOT(tc, f) {
         ctx = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTContext);
         MVM_ASSIGN_REF(tc, &(ctx->header), ((MVMContext *)ctx)->body.context, f);
-    });
+        ((MVMContext *)ctx)->body.traversable = 1;
+    }
+    return ctx;
+}
+
+/* Creates a MVMContent wrapper object around an MVMFrame that only allows
+ * access to the current frame and does not support traversal. */
+MVMObject * MVM_context_from_frame_non_traversable(MVMThreadContext *tc, MVMFrame *f) {
+    MVMObject *ctx;
+    f = MVM_frame_force_to_heap(tc, f);
+    MVMROOT(tc, f) {
+        ctx = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTContext);
+        MVM_ASSIGN_REF(tc, &(ctx->header), ((MVMContext *)ctx)->body.context, f);
+    }
     return ctx;
 }
 
@@ -315,6 +332,11 @@ static MVMint32 traversal_exists(MVMThreadContext *tc, MVMFrame *base, MVMuint8
 /* Try to get a context with the specified traversal applied. Ensures that the
  * traversal would lead to a result, and returns a NULL context if not. */
 MVMObject * MVM_context_apply_traversal(MVMThreadContext *tc, MVMContext *ctx, MVMuint8 traversal) {
+    /* Ensure we've got a traversable context. */
+    if (!ctx->body.traversable)
+        MVM_exception_throw_adhoc(tc,
+                "Cannot move to outers or callers with non-traversable context");
+
     /* Create new traversals array with the latest one at the end. */
     MVMuint32 new_num_traversals = ctx->body.num_traversals + 1;
     MVMuint8 *new_traversals = MVM_malloc(new_num_traversals);
@@ -326,12 +348,13 @@ MVMObject * MVM_context_apply_traversal(MVMThreadContext *tc, MVMContext *ctx, M
     if (traversal_exists(tc, ctx->body.context, new_traversals, new_num_traversals)) {
         /* Yes, make a new context object and return it. */
         MVMContext *result;
-        MVMROOT(tc, ctx, {
+        MVMROOT(tc, ctx) {
             result = (MVMContext *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTContext);
-        });
+        }
         MVM_ASSIGN_REF(tc, &(result->common.header), result->body.context, ctx->body.context);
         result->body.traversals = new_traversals;
         result->body.num_traversals = new_num_traversals;
+        result->body.traversable = 1;
         return (MVMObject *)result;
     }
     else {
@@ -416,7 +439,7 @@ MVMObject * MVM_context_lexical_lookup(MVMThreadContext *tc, MVMContext *ctx, MV
     MVMSpeshFrameWalker fw;
     MVM_spesh_frame_walker_init_for_outers(tc, &fw, ctx->body.context);
     if (apply_traversals(tc, &fw, ctx->body.traversals, ctx->body.num_traversals)) {
-        MVMRegister *result = MVM_frame_lexical_lookup_using_frame_walker(tc, &fw, name);
+        MVMRegister *result = MVM_frame_lexical_lookup_using_frame_walker(tc, &fw, name, MVM_reg_obj);
         return result ? result->o : tc->instance->VMNull;
     }
     MVM_spesh_frame_walker_cleanup(tc, &fw);
@@ -440,7 +463,7 @@ MVMObject * MVM_context_caller_lookup(MVMThreadContext *tc, MVMContext *ctx, MVM
     MVMSpeshFrameWalker fw;
     MVM_spesh_frame_walker_init(tc, &fw, ctx->body.context, 1);
     if (apply_traversals(tc, &fw, ctx->body.traversals, ctx->body.num_traversals)) {
-        MVMRegister *result = MVM_frame_lexical_lookup_using_frame_walker(tc, &fw, name);
+        MVMRegister *result = MVM_frame_lexical_lookup_using_frame_walker(tc, &fw, name, MVM_reg_obj);
         return result ? result->o : tc->instance->VMNull;
     }
     MVM_spesh_frame_walker_cleanup(tc, &fw);
