@@ -486,9 +486,9 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
      * shall need, since later we may allocate a result and and move it. */
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
     if (MVM_UNLIKELY(!body->lib_handle)) {
-        MVMROOT3(tc, site, args, res_type, {
+        MVMROOT3(tc, site, args, res_type) {
             MVM_nativecall_restore_library(tc, body, site);
-        });
+        }
         body = MVM_nativecall_get_nc_body(tc, site);
     }
     MVMint16  num_args    = body->num_args;
@@ -616,7 +616,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     }
 
 
-    MVMROOT3(tc, args, res_type, result, {
+    MVMROOT3(tc, args, res_type, result) {
         MVM_gc_mark_thread_blocked(tc);
         if (result) {
             /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
@@ -689,7 +689,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                     break;
                 case MVM_NATIVECALL_ARG_CALLBACK: {
                     /* TODO: A callback -return- value means that we have a C method
-                    * that needs to be wrapped similarly to a is native(...) Perl 6
+                    * that needs to be wrapped similarly to a is native(...) Rakudo
                     * sub. */
                     void *ret;
                     ffi_call(&cif, entry_point, &ret, values);
@@ -719,7 +719,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                     MVM_exception_throw_adhoc(tc, "Internal error: unhandled libffi return type");
             }
         }
-    });
+    }
 
     for (i = 0; i < num_args; i++) {
         MVMObject *value = MVM_repr_at_pos_o(tc, args, i);
@@ -859,9 +859,9 @@ void MVM_nativecall_dispatch(MVMThreadContext *tc, MVMObject *res_type,
      * shall need, since later we may allocate a result and and move it. */
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
     if (MVM_UNLIKELY(!body->lib_handle)) {
-        MVMROOT2(tc, site, res_type, {
+        MVMROOT2(tc, site, res_type) {
             MVM_nativecall_restore_library(tc, body, site);
-        });
+        }
         body = MVM_nativecall_get_nc_body(tc, site);
     }
     MVMint16  num_args    = body->num_args;
@@ -1103,7 +1103,7 @@ void MVM_nativecall_dispatch(MVMThreadContext *tc, MVMObject *res_type,
     }
 
 
-    MVMROOT2(tc, res_type, result, {
+    MVMROOT2(tc, res_type, result) {
         MVM_gc_mark_thread_blocked(tc);
         if (result) {
             /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
@@ -1209,7 +1209,7 @@ void MVM_nativecall_dispatch(MVMThreadContext *tc, MVMObject *res_type,
                     break;
                 case MVM_NATIVECALL_ARG_CALLBACK: {
                     /* TODO: A callback -return- value means that we have a C method
-                    * that needs to be wrapped similarly to a is native(...) Perl 6
+                    * that needs to be wrapped similarly to a is native(...) Rakudo
                     * sub. */
                     void *ret;
                     ffi_call(&cif, entry_point, &ret, values);
@@ -1242,7 +1242,7 @@ void MVM_nativecall_dispatch(MVMThreadContext *tc, MVMObject *res_type,
                         "in MVM_nativecall_dispatch", ret_type & MVM_NATIVECALL_ARG_TYPE_MASK);
             }
         }
-    });
+    }
 
     /* Free any memory that we need to. */
     if (free_strs)
