@@ -1,5 +1,4 @@
 #include "moar.h"
-#include <platform/threads.h>
 #include "platform/malloc_trim.h"
 
 /* If we have the job of doing GC for a thread, we add it to our work
@@ -30,7 +29,7 @@ static MVMuint32 signal_one_thread(MVMThreadContext *tc, MVMThreadContext *to_si
     unsigned int had_suspend_request = 0;
     while (1) {
         AO_t current = MVM_load(&to_signal->gc_status);
-        switch (current) {
+        switch (AO_READ(current)) {
             case MVMGCStatus_NONE:
                 /* Try to set it from running to interrupted - the common case. */
                 if (MVM_cas(&to_signal->gc_status, MVMGCStatus_NONE,
@@ -188,8 +187,7 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
         MVM_profile_heap_take_snapshot(tc);
 
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
-            "Thread %d run %d : Co-ordinator handling fixed-size allocator safepoint frees\n");
-        MVM_fixed_size_safepoint(tc, tc->instance->fsa);
+            "Thread %d run %d : Co-ordinator handling allocator safepoint frees\n");
         MVM_alloc_safepoint(tc);
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator signalling in-trays clear\n");
@@ -356,7 +354,8 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
                     }
                 }
             } else if (MVM_load(&tc->gc_status) == MVMGCStatus_NONE) {
-                fprintf(stderr, "marking thread %d unblocked, but its status is already NONE.\n", tc->thread_id);
+                if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+                    fprintf(stderr, "marking thread %d unblocked, but its status is already NONE.\n", tc->thread_id);
                 break;
             } else {
                 MVM_platform_thread_yield();
@@ -477,7 +476,8 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
         uv_mutex_lock(&tc->instance->mutex_threads);
         cur_thread = tc->instance->threads;
         while (cur_thread) {
-            data[8] += cur_thread->body.tc->num_gen2roots;
+            if (cur_thread->body.tc)
+                data[8] += cur_thread->body.tc->num_gen2roots;
             cur_thread = cur_thread->body.next;
         }
         uv_mutex_unlock(&tc->instance->mutex_threads);
@@ -624,6 +624,46 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
  * Those cases can be distinguished by the gc state masked with
  * MVMSUSPENDSTATUS_MASK.
  *   */
+static int react_to_debugserver_request(MVMThreadContext *tc) {
+    if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+        fprintf(stderr, "thread %p has received a request.\n", tc);
+
+    MVMDebugServerRequestKind kind = tc->instance->debugserver->request_data.kind;
+
+    if (kind == MVM_DebugRequest_invoke) {
+        MVMCode *invoke_target = tc->instance->debugserver->request_data.data.invoke.target;
+        MVMArgs *args = tc->instance->debugserver->request_data.data.invoke.args;
+        tc->instance->debugserver->request_data.data.invoke.target = NULL;
+
+        /* Have to clear the threadcontext's "blocked" status so that invoke
+         * can work, for example to run the instruction level barrier.
+         */
+
+        if (!MVM_trycas(&tc->gc_status, MVMGCStatus_UNABLE | MVMSuspendState_SUSPENDED, MVMGCStatus_NONE)) {
+            MVM_panic(MVM_exitcode_gcorch, "could not unblock/unsuspend thread");
+        }
+
+        MVM_frame_dispatch(tc, invoke_target, *args, -1);
+
+        MVM_gc_mark_thread_blocked(tc);
+    }
+    else {
+        if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+            fprintf(stderr, "this debug request kind not implemented: %d\n", kind);
+        return 0;
+    }
+
+    if (MVM_cas(&tc->instance->debugserver->request_data.status,
+                MVM_DebugRequestStatus_sender_is_waiting,
+                MVM_DebugRequestStatus_receiver_acknowledged)
+        != MVM_DebugRequestStatus_sender_is_waiting) {
+        if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+            fprintf(stderr, "could not acknowledge request?!?\n");
+    }
+    tc->instance->debugserver->request_data.kind = MVM_DebugRequest_empty;
+
+    return 1;
+}
 void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
     GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Entered from interrupt\n");
 
@@ -633,14 +673,21 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
             fprintf(stderr, "thread %d reacting to suspend request\n", tc->thread_id);
         MVM_gc_mark_thread_blocked(tc);
         while (1) {
+            uv_mutex_lock(&tc->instance->debugserver->mutex_cond);
             uv_cond_wait(&tc->instance->debugserver->tell_threads, &tc->instance->debugserver->mutex_cond);
+            uv_mutex_unlock(&tc->instance->debugserver->mutex_cond);
             if ((MVM_load(&tc->gc_status) & MVMSUSPENDSTATUS_MASK) == MVMSuspendState_NONE) {
                 if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
                     fprintf(stderr, "thread %d got un-suspended\n", tc->thread_id);
                 break;
             } else {
+                if (tc->instance->debugserver && tc->instance->debugserver->request_data.target_tc == tc) {
+                    if (react_to_debugserver_request(tc)) {
+                        break;
+                    }
+                }
                 if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
-                    fprintf(stderr, "something happened, but we're still suspended.\n");
+                    fprintf(stderr, "thread %p: something happened, but we're still suspended.\n", tc);
             }
         }
         MVM_gc_mark_thread_unblocked(tc);
