@@ -251,9 +251,9 @@ static void two_complement_shl(MVMThreadContext *tc, mp_int *result, mp_int *val
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *source) { \
     MVMP6bigintBody *bb; \
     MVMObject *result; \
-    MVMROOT(tc, source, { \
+    MVMROOT(tc, source) { \
         result = MVM_repr_alloc_init(tc, result_type);\
-    }); \
+    } \
     bb = get_bigint_body(tc, result); \
     if (!IS_CONCRETE(source)) { \
         store_int64_result(tc, bb, 0); \
@@ -292,9 +292,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     MVMObject *result; \
     mp_err err; \
     mp_int *ia, *ib, *ic; \
-    MVMROOT2(tc, a, b, { \
+    MVMROOT2(tc, a, b) { \
         result = MVM_repr_alloc_init(tc, result_type);\
-    }); \
+    } \
     ba = get_bigint_body(tc, a); \
     bb = get_bigint_body(tc, b); \
     bc = get_bigint_body(tc, result); \
@@ -343,9 +343,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
         mp_err err; \
         mp_int *ia, *ib, *ic; \
-        MVMROOT2(tc, a, b, { \
+        MVMROOT2(tc, a, b) { \
             result = MVM_repr_alloc_init(tc, result_type);\
-        }); \
+        } \
         ba = get_bigint_body(tc, a); \
         bb = get_bigint_body(tc, b); \
         bc = get_bigint_body(tc, result); \
@@ -382,9 +382,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
 #define MVM_BIGINT_BINARY_OP_2(opname, SMALLINT_OP) \
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) { \
     MVMObject *result; \
-    MVMROOT2(tc, a, b, { \
+    MVMROOT2(tc, a, b) { \
         result = MVM_repr_alloc_init(tc, result_type);\
-    }); \
+    } \
     {\
         MVMP6bigintBody *ba = get_bigint_body(tc, a); \
         MVMP6bigintBody *bb = get_bigint_body(tc, b); \
@@ -427,9 +427,9 @@ MVM_BIGINT_BINARY_OP(lcm)
 MVMObject *MVM_bigint_gcd(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
     MVMObject       *result;
 
-    MVMROOT2(tc, a, b, {
+    MVMROOT2(tc, a, b) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     {
         MVMP6bigintBody *ba = get_bigint_body(tc, a);
@@ -494,9 +494,9 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
 
     MVMObject *result;
 
-    MVMROOT2(tc, a, b, {
+    MVMROOT2(tc, a, b) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     {
         MVMP6bigintBody *ba = get_bigint_body(tc, a);
@@ -549,9 +549,9 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         return a;
     }
 
-    MVMROOT2(tc, a, b, {
+    MVMROOT2(tc, a, b) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, b);
@@ -707,9 +707,9 @@ MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
-    MVMROOT(tc, a, {
+    MVMROOT(tc, a) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
@@ -752,9 +752,9 @@ MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
-    MVMROOT(tc, a, {
+    MVMROOT(tc, a) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
@@ -786,9 +786,9 @@ MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
-    MVMROOT(tc, a, {
+    MVMROOT(tc, a) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
@@ -840,9 +840,9 @@ MVMObject *MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result_type, MVMOb
         MVM_exception_throw_adhoc(tc, "Error creating a big integer: %s", mp_error_to_string(err));
     }
 
-    MVMROOT3(tc, a, b, c, {
+    MVMROOT3(tc, a, b, c) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     bd = get_bigint_body(tc, result);
 
@@ -886,9 +886,9 @@ MVMObject * MVM_coerce_sI(MVMThreadContext *tc, MVMString *s, MVMObject *type) {
     int is_malloced = 0;
     MVMStringIndex i;
     MVMObject *a;
-    MVMROOT(tc, s, {
+    MVMROOT(tc, s) {
         a = MVM_repr_alloc_init(tc, type);
-    });
+    }
     if (s->body.num_graphs < 120) {
         buf = alloca(s->body.num_graphs + 1);
     }
@@ -903,6 +903,9 @@ MVMObject * MVM_coerce_sI(MVMThreadContext *tc, MVMString *s, MVMObject *type) {
         case MVM_STRING_GRAPHEME_8:
             memcpy(buf, s->body.storage.blob_8, sizeof(MVMGrapheme8) * s->body.num_graphs);
             break;
+        case MVM_STRING_IN_SITU_8:
+            memcpy(buf, s->body.storage.in_situ_8, sizeof(MVMGrapheme8) * s->body.num_graphs);
+            break;
         case MVM_STRING_GRAPHEME_32:
             for (i = 0; i < s->body.num_graphs; i++) {
                 buf[i] = can_fit_into_8bit(s->body.storage.blob_32[i])
@@ -910,6 +913,13 @@ MVMObject * MVM_coerce_sI(MVMThreadContext *tc, MVMString *s, MVMObject *type) {
                     : '?'; /* Add a filler bogus char if it can't fit */
             }
             break;
+        case MVM_STRING_IN_SITU_32:
+            for (i = 0; i < s->body.num_graphs; i++) {
+                buf[i] = can_fit_into_8bit(s->body.storage.in_situ_32[i])
+                    ? s->body.storage.in_situ_32[i]
+                    : '?'; /* Add a filler bogus char if it can't fit */
+            }
+            break;
         case MVM_STRING_STRAND: {
             MVMGraphemeIter gi;
             MVM_string_gi_init(tc, &gi, s);
@@ -935,9 +945,9 @@ MVMObject * MVM_bigint_from_bigint(MVMThreadContext *tc, MVMObject *result_type,
     MVMP6bigintBody *r_body;
     MVMObject       *result;
 
-    MVMROOT(tc, a, {
+    MVMROOT(tc, a) {
         result = MVM_repr_alloc_init(tc, result_type);
-    });
+    }
 
     a_body = get_bigint_body(tc, a);
     r_body = get_bigint_body(tc, result);
@@ -1464,7 +1474,7 @@ mp_err MVM_mp_rand(MVMThreadContext *tc, mp_int *a, int digits)
 }
 
 
-/* 
+/*
     The old version of LibTomMath has it publically defined the new one not,
     so we can take the (non)existance as a marker.
  */
@@ -1500,9 +1510,9 @@ MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b)
             if (have_to_negate)
                 result_int *= -1;
 
-            MVMROOT2(tc, type, b, {
+            MVMROOT2(tc, type, b) {
                 result = MVM_repr_alloc_init(tc, type);
-            });
+            }
 
             ba = get_bigint_body(tc, result);
             store_int64_result(tc, ba, result_int);
@@ -1516,9 +1526,9 @@ MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b)
         mp_int *rnd = MVM_malloc(sizeof(mp_int));
         mp_int *max = force_bigint(tc, bb, 0);
 
-        MVMROOT2(tc, type, b, {
+        MVMROOT2(tc, type, b) {
             result = MVM_repr_alloc_init(tc, type);
-        });
+        }
 
         ba = get_bigint_body(tc, result);
 
