@@ -4,9 +4,6 @@
  * optionally visiting its static chain at each point, and getting correct
  * results even if inlining has taken place.*/
 
-/* Sentinel value to indicate there's no inline to explore. */
-#define NO_INLINE -2
-
 /* Initializes the frame walker. The `MVMSpeshFrameWalker` object MUST be on
  * the system stack, and the cleanup function MUST be called after using it,
  * except in the case of an exception. This is because, since frames are GC
@@ -18,7 +15,7 @@ static void init_common(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MVMFrame
     fw->cur_outer_frame = NULL;
     fw->started = fw->traversed = 0;
     fw->visiting_outers = 0;
-    fw->inline_idx = NO_INLINE;
+    fw->inline_idx = MVM_SPESH_FRAME_WALKER_NO_INLINE;
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&(fw->cur_caller_frame));
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&(fw->cur_outer_frame));
 }
@@ -41,11 +38,11 @@ void MVM_spesh_frame_walker_init_for_outers(MVMThreadContext *tc, MVMSpeshFrameW
 /* Go to the next inline, if any. */
 static void go_to_next_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     MVMFrame *f = fw->cur_caller_frame;
-    if (fw->inline_idx == NO_INLINE)
+    if (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE)
         return;
     MVMSpeshCandidate *cand = f->spesh_cand;
     if (!cand) {
-        fw->inline_idx = NO_INLINE;
+        fw->inline_idx = MVM_SPESH_FRAME_WALKER_NO_INLINE;
         return;
     }
     MVMJitCode *jitcode = cand->body.jitcode;
@@ -68,19 +65,22 @@ static void go_to_next_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     }
 
     /* No inline available. */
-    fw->inline_idx = NO_INLINE;
+    fw->inline_idx = MVM_SPESH_FRAME_WALKER_NO_INLINE;
 }
 
 /* See if the current frame is specialized, and if so if we are in an inline.
  * If so, go to the innermost inline. */
 static void go_to_first_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MVMFrame *prev) {
     MVMFrame *f = fw->cur_caller_frame;
+    /* Get a local copy of spesh_cand so we don't stumble over another thread
+     * clearing the pointer after we checked it */
     MVMSpeshCandidate *spesh_cand = f->spesh_cand;
     if (spesh_cand && spesh_cand->body.inlines) {
         MVMJitCode *jitcode = spesh_cand->body.jitcode;
         if (jitcode && f->jit_entry_label) {
-            void *current_position = prev && prev->extra && prev->extra->caller_jit_position
-                ? prev->extra->caller_jit_position
+            MVMFrameExtra *extra; /* Get a local copy as well for the same reason */
+            void *current_position = prev && (extra = prev->extra) && extra->caller_jit_position
+                ? extra->caller_jit_position
                 : MVM_jit_code_get_current_position(tc, jitcode, f);
             MVMuint32 idx = MVM_jit_code_get_active_inlines(tc, jitcode, current_position, 0);
             if (idx < jitcode->num_inlines) {
@@ -90,29 +90,38 @@ static void go_to_first_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MV
             }
         }
         else {
-            MVMint32 deopt_idx = prev && prev->extra && prev->extra->caller_deopt_idx > 0
-                ? prev->extra->caller_deopt_idx - 1
-                : MVM_spesh_deopt_find_inactive_frame_deopt_idx(tc, f);
-            if (deopt_idx >= 0) {
-                fw->deopt_offset = spesh_cand->body.deopts[2 * deopt_idx + 1];
+            if (!fw->started && f == tc->cur_frame) {
+                fw->deopt_offset = *(tc->interp_cur_op) - spesh_cand->body.bytecode;
                 fw->inline_idx = -1;
                 go_to_next_inline(tc, fw);
                 return;
             }
+            else {
+                MVMFrameExtra *extra; /* Get a local copy as well for the same reason */
+                MVMint32 deopt_idx = prev && (extra = prev->extra) && extra->caller_deopt_idx > 0
+                    ? extra->caller_deopt_idx - 1
+                    : MVM_spesh_deopt_find_inactive_frame_deopt_idx(tc, f, spesh_cand);
+                if (deopt_idx >= 0) {
+                    fw->deopt_offset = MVM_spesh_deopt_bytecode_pos(spesh_cand->body.deopts[2 * deopt_idx + 1]);
+                    fw->inline_idx = -1;
+                    go_to_next_inline(tc, fw);
+                    return;
+                }
+            }
         }
     }
-    fw->inline_idx = NO_INLINE;
+    fw->inline_idx = MVM_SPESH_FRAME_WALKER_NO_INLINE;
 }
 
 /* Moves one caller frame deeper, accounting for inlines. */
-MVMuint32 move_one_caller(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
+static MVMuint32 move_one_caller(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     MVMFrame *caller;
 
      /* Is there an inline to try and visit? If there is one, then
      * we will be placed on it. If there is not one, we will be placed
      * on the base frame containing inlines. Either way, we must have
      * something to go to. */
-    if (fw->inline_idx != NO_INLINE) {
+    if (fw->inline_idx != MVM_SPESH_FRAME_WALKER_NO_INLINE) {
         go_to_next_inline(tc, fw);
         return 1;
     }
@@ -136,8 +145,8 @@ MVMuint32 move_one_caller(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
  * over). */
 MVMuint32 MVM_spesh_frame_walker_next(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     if (!fw->started) {
-        fw->started = 1;
         go_to_first_inline(tc, fw, NULL);
+        fw->started = 1;
         return fw->cur_caller_frame ? 1 : 0;
     }
     else if (fw->traversed) {
@@ -165,7 +174,7 @@ MVMuint32 MVM_spesh_frame_walker_next(MVMThreadContext *tc, MVMSpeshFrameWalker
         else if (fw->visit_outers) {
             MVMFrame *outer;
             MVMSpeshCandidate *spesh_cand = fw->cur_caller_frame->spesh_cand;
-            if (fw->inline_idx == NO_INLINE || !spesh_cand) {
+            if (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE || !spesh_cand) {
                 outer = fw->cur_caller_frame->outer;
             }
             else {
@@ -186,19 +195,6 @@ MVMuint32 MVM_spesh_frame_walker_next(MVMThreadContext *tc, MVMSpeshFrameWalker
     }
 }
 
-/* Returns non-zero if we're currently visiting an inline, zero otherwise. */
-MVMuint32 MVM_spesh_frame_walker_is_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
-    return fw->inline_idx != NO_INLINE;
-}
-
-/* Gets the current frame we're walking. If we're in an inline, then it's the
- * frame holding the inline. */
-MVMFrame * MVM_spesh_frame_walker_current_frame(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
-    return MVM_UNLIKELY(fw->visiting_outers)
-        ? fw->cur_outer_frame
-        : fw->cur_caller_frame;
-}
-
 static void find_lex_info(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MVMFrame **cur_frame_out,
                           MVMStaticFrame **sf_out, MVMuint32 *base_index_out) {
     if (fw->visiting_outers) {
@@ -209,7 +205,7 @@ static void find_lex_info(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MVMFram
     else {
         *cur_frame_out = fw->cur_caller_frame;
         MVMSpeshCandidate *spesh_cand = fw->cur_caller_frame->spesh_cand;
-        if (fw->inline_idx == NO_INLINE || !spesh_cand) {
+        if (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE || !spesh_cand) {
             *sf_out = fw->cur_caller_frame->static_info;
             *base_index_out = 0;
         }
@@ -232,6 +228,8 @@ MVMuint32 MVM_spesh_frame_walker_get_lex(MVMThreadContext *tc, MVMSpeshFrameWalk
     MVMStaticFrame *sf;
     MVMuint32 base_index;
     find_lex_info(tc, fw, &cur_frame, &sf, &base_index);
+    if (sf->body.num_lexicals == 0)
+        return 0;
     MVMuint32 idx = MVM_get_lexical_by_name(tc, sf, name);
     if (idx != MVM_INDEX_HASH_NOT_FOUND) {
         MVMint32 index = base_index + idx;
@@ -240,9 +238,9 @@ MVMuint32 MVM_spesh_frame_walker_get_lex(MVMThreadContext *tc, MVMSpeshFrameWalk
         *found_out = result;
         *found_kind_out = kind;
         if (vivify && kind == MVM_reg_obj && !result->o) {
-            MVMROOT(tc, cur_frame, {
+            MVMROOT(tc, cur_frame) {
                     MVM_frame_vivify_lexical(tc, cur_frame, index);
-                });
+                }
         }
         if (found_frame)
             *found_frame = cur_frame;
@@ -255,7 +253,7 @@ MVMuint32 MVM_spesh_frame_walker_get_lex(MVMThreadContext *tc, MVMSpeshFrameWalk
 MVMuint32 MVM_spesh_frame_walker_move_outer(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     MVMFrame *outer;
     MVMSpeshCandidate *spesh_cand = fw->cur_caller_frame->spesh_cand;
-    if (fw->inline_idx == NO_INLINE || !spesh_cand) {
+    if (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE || !spesh_cand) {
         outer = fw->cur_caller_frame->outer;
     }
     else {
@@ -265,7 +263,7 @@ MVMuint32 MVM_spesh_frame_walker_move_outer(MVMThreadContext *tc, MVMSpeshFrameW
     }
     fw->cur_caller_frame = outer;
     fw->cur_outer_frame = NULL;
-    fw->inline_idx = NO_INLINE;
+    fw->inline_idx = MVM_SPESH_FRAME_WALKER_NO_INLINE;
     fw->visiting_outers = 0;
     fw->started = 1;
     if (outer != NULL) {
@@ -304,7 +302,7 @@ MVMuint32 MVM_spesh_frame_walker_move_caller_skip_thunks(MVMThreadContext *tc,
                                                          MVMSpeshFrameWalker *fw) {
     while (MVM_spesh_frame_walker_move_caller(tc, fw)) {
         MVMSpeshCandidate *spesh_cand = fw->cur_caller_frame->spesh_cand;
-        MVMStaticFrame *sf = (fw->inline_idx == NO_INLINE || !spesh_cand)
+        MVMStaticFrame *sf = (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE || !spesh_cand)
             ? fw->cur_caller_frame->static_info
             : spesh_cand->body.inlines[fw->inline_idx].sf;
         if (!sf->body.is_thunk)
@@ -318,7 +316,7 @@ MVMuint32 MVM_spesh_frame_walker_move_caller_skip_thunks(MVMThreadContext *tc,
 MVMFrame * MVM_spesh_frame_walker_get_frame(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     if (fw->visiting_outers)
         return fw->cur_outer_frame;
-    if (fw->inline_idx == NO_INLINE)
+    if (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE)
         return fw->cur_caller_frame;
     return NULL;
 }
@@ -331,7 +329,7 @@ MVMObject * MVM_spesh_frame_walker_get_lexicals_hash(MVMThreadContext *tc, MVMSp
     MVMHLLConfig *hll = MVM_hll_current(tc);
     MVMObject *ctx_hash = MVM_repr_alloc_init(tc, hll->slurpy_hash_type);
     find_lex_info(tc, fw, &frame, &sf, &base_index);
-    MVMROOT3(tc, ctx_hash, frame, sf, {
+    MVMROOT3(tc, ctx_hash, frame, sf) {
         MVMString **lexnames = sf->body.lexical_names_list;
         MVMuint32 i;
         for (i = 0; i < sf->body.num_lexicals; i++) {
@@ -417,7 +415,7 @@ MVMObject * MVM_spesh_frame_walker_get_lexicals_hash(MVMThreadContext *tc, MVMSp
                             MVM_reg_get_debug_name(tc, type));
             }
         }
-    });
+    }
     return ctx_hash;
 }
 
@@ -429,6 +427,8 @@ MVMint64 MVM_spesh_frame_walker_get_lexical_primspec(MVMThreadContext *tc,
     MVMStaticFrame *sf;
     MVMuint32 base_index;
     find_lex_info(tc, fw, &cur_frame, &sf, &base_index);
+    if (sf->body.num_lexicals == 0)
+        return -1;
     MVMuint32 idx = MVM_get_lexical_by_name(tc, sf, name);
     if (idx != MVM_INDEX_HASH_NOT_FOUND)
         return MVM_frame_translate_to_primspec(tc, sf->body.lexical_types[idx]);
@@ -440,7 +440,7 @@ MVMObject * MVM_spesh_frame_walker_get_code(MVMThreadContext *tc, MVMSpeshFrameW
     if (fw->visiting_outers)
         return fw->cur_outer_frame->code_ref;
     MVMSpeshCandidate *spesh_cand = fw->cur_caller_frame->spesh_cand;
-    if (fw->inline_idx == NO_INLINE || !spesh_cand)
+    if (fw->inline_idx == MVM_SPESH_FRAME_WALKER_NO_INLINE || !spesh_cand)
         return fw->cur_caller_frame->code_ref;
     return fw->cur_caller_frame->work[
         spesh_cand->body.inlines[fw->inline_idx].code_ref_reg
@@ -456,8 +456,3 @@ MVMuint64 MVM_spesh_frame_walker_get_lexical_count(MVMThreadContext *tc, MVMSpes
     find_lex_info(tc, fw, &cur_frame, &sf, &base_index);
     return sf->body.num_lexicals;
 }
-
-/* Cleans up the spesh frame walker after use. */
-void MVM_spesh_frame_walker_cleanup(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
-    MVM_gc_root_temp_pop_n(tc, 2);
-}
