@@ -23,6 +23,12 @@ MVM_STATIC_INLINE MVMuint32 GET_UI32(const MVMuint8 *pc, MVMint32 idx) {
     return retval;
 }
 
+MVM_STATIC_INLINE MVMuint64 GET_UI64(const MVMuint8 *pc, MVMint32 idx) {
+    MVMuint64 retval;
+    memcpy(&retval, pc + idx, sizeof(retval));
+    return retval;
+}
+
 MVM_STATIC_INLINE MVMuint32 GET_N32(const MVMuint8 *pc, MVMint32 idx) {
     MVMnum32 retval;
     memcpy(&retval, pc + idx, sizeof(retval));
@@ -35,19 +41,6 @@ void * MVM_spesh_alloc(MVMThreadContext *tc, MVMSpeshGraph *g, size_t bytes) {
     return MVM_region_alloc(tc, &g->region_alloc, bytes);
 }
 
-/* Looks up op info; doesn't sanity check, since we should be working on code
- * that already pass validation. */
-static const MVMOpInfo * get_op_info(MVMThreadContext *tc, MVMCompUnit *cu, MVMuint16 opcode) {
-    if (opcode < MVM_OP_EXT_BASE) {
-        return MVM_op_get_op(opcode);
-    }
-    else {
-        MVMuint16       index  = opcode - MVM_OP_EXT_BASE;
-        MVMExtOpRecord *record = &cu->body.extops[index];
-        return MVM_ext_resolve_extop_record(tc, record);
-    }
-}
-
 /* Grows the spesh graph's deopt table if it is already full, so that we have
  * space for 1 more entry. */
 void MVM_spesh_graph_grow_deopt_table(MVMThreadContext *tc, MVMSpeshGraph *g) {
@@ -121,6 +114,17 @@ static void add_logged_annotation(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpe
     ins_node->annotations = ann;
 }
 
+/* Records the current bytecode position as an inline cache annotation. Used for
+ * resolving to the correct inline cache entry. */
+static void add_cached_annotation(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins_node,
+                                  MVMuint8 *pc) {
+    MVMSpeshAnn *ann = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
+    ann->type = MVM_SPESH_ANN_CACHED;
+    ann->data.bytecode_offset = pc - g->bytecode;
+    ann->next = ins_node->annotations;
+    ins_node->annotations = ann;
+}
+
 /* Finds the linearly previous basic block (not cheap, but uncommon). */
 MVMSpeshBB * MVM_spesh_graph_linear_prev(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *search) {
     MVMSpeshBB *bb = g->entry;
@@ -146,6 +150,62 @@ static MVMint32 already_succs(MVMThreadContext *tc, MVMSpeshBB *bb, MVMSpeshBB *
     return 0;
 }
 
+/* Checks if the op is one of the spesh dispatch ops. */
+static MVMint32 spesh_dispatchy(MVMuint16 opcode) {
+    return (opcode >= MVM_OP_sp_dispatch_v    && opcode <= MVM_OP_sp_dispatch_o)    ||
+           (opcode >= MVM_OP_sp_runbytecode_v && opcode <= MVM_OP_sp_runbytecode_o) ||
+           (opcode >= MVM_OP_sp_runcfunc_v    && opcode <= MVM_OP_sp_runcfunc_o) ||
+           (opcode >= MVM_OP_sp_runnativecall_v && opcode <= MVM_OP_sp_runnativecall_o);
+}
+
+int MVM_spesh_graph_ins_ends_bb(MVMThreadContext *tc, const MVMOpInfo *info) {
+    switch (info->opcode) {
+    case MVM_OP_return_i:
+    case MVM_OP_return_u:
+    case MVM_OP_return_n:
+    case MVM_OP_return_s:
+    case MVM_OP_return_o:
+    case MVM_OP_return:
+    case MVM_OP_dispatch_v:
+    case MVM_OP_dispatch_i:
+    case MVM_OP_dispatch_u:
+    case MVM_OP_dispatch_n:
+    case MVM_OP_dispatch_s:
+    case MVM_OP_dispatch_o:
+    case MVM_OP_sp_dispatch_v:
+    case MVM_OP_sp_dispatch_i:
+    case MVM_OP_sp_dispatch_u:
+    case MVM_OP_sp_dispatch_n:
+    case MVM_OP_sp_dispatch_s:
+    case MVM_OP_sp_dispatch_o:
+    case MVM_OP_sp_runbytecode_v:
+    case MVM_OP_sp_runbytecode_i:
+    case MVM_OP_sp_runbytecode_u:
+    case MVM_OP_sp_runbytecode_n:
+    case MVM_OP_sp_runbytecode_s:
+    case MVM_OP_sp_runbytecode_o:
+    case MVM_OP_sp_runcfunc_v:
+    case MVM_OP_sp_runcfunc_i:
+    case MVM_OP_sp_runcfunc_u:
+    case MVM_OP_sp_runcfunc_n:
+    case MVM_OP_sp_runcfunc_s:
+    case MVM_OP_sp_runcfunc_o:
+    case MVM_OP_sp_runnativecall_v:
+    case MVM_OP_sp_runnativecall_i:
+    case MVM_OP_sp_runnativecall_u:
+    case MVM_OP_sp_runnativecall_n:
+    case MVM_OP_sp_runnativecall_s:
+    case MVM_OP_sp_runnativecall_o:
+        return 1;
+    default:
+        if (info->jittivity & (MVM_JIT_INFO_THROWISH | MVM_JIT_INFO_INVOKISH)) {
+            return 1;
+        }
+        break;
+    }
+    return 0;
+}
+
 /* Builds the control flow graph, populating the passed spesh graph structure
  * with it. This also makes nodes for all of the instruction. */
 #define MVM_CFG_BB_START    1
@@ -207,8 +267,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         /* Look up op info. */
         MVMuint16  opcode     = *(MVMuint16 *)pc;
         MVMuint8  *args       = pc + 2;
-        MVMuint8   arg_size   = 0;
-        const MVMOpInfo *info = get_op_info(tc, cu, opcode);
+        MVMuint16  arg_size   = 0;
+        const MVMOpInfo *info = MVM_bytecode_get_validated_op_info(tc, cu, opcode);
 
         /* Create an instruction node, add it, and record its position. */
         MVMSpeshIns *ins_node = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
@@ -238,7 +298,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         /* If this is a pre-instruction deopt point opcode, annotate. */
         if (!existing_deopts && (info->deopt_point & MVM_DEOPT_MARK_ONE_PRE))
             MVM_spesh_graph_add_deopt_annotation(tc, g, ins_node,
-                pc - g->bytecode, MVM_SPESH_ANN_DEOPT_ONE_INS);
+                pc - g->bytecode, MVM_SPESH_ANN_DEOPT_PRE_INS);
 
         /* Let's see if we have a line-number annotation. */
         if (ann_ptr && pc - sf->body.bytecode == ann_ptr->bytecode_offset) {
@@ -252,6 +312,29 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             MVM_bytecode_advance_annotation(tc, &sf->body, ann_ptr);
         }
 
+        /* Dispatch ops have a variable number of arguments based on a callsite. */
+        if (MVM_op_get_mark(info->opcode)[1] == 'd' ||
+                (MVM_op_get_mark(info->opcode)[1] == 's' && spesh_dispatchy(info->opcode))) {
+            /* Fake up an op info struct for the duration of this spesh graph's
+             * life, so everything else can pretend it's got a fixed number of
+             * arguments. */
+            MVMCallsite *callsite = MVM_spesh_disp_callsite_for_dispatch_op(opcode, args, cu);
+            ins_node->info = MVM_spesh_alloc(tc, g, MVM_spesh_disp_dispatch_op_info_size(
+                tc, info, callsite));
+            MVM_spesh_disp_initialize_dispatch_op_info(tc, info, callsite, (MVMOpInfo*)ins_node->info);
+            info = ins_node->info;
+        }
+
+        /* The sp_resumption op is also var args. */
+        else if (info->opcode == MVM_OP_sp_resumption) {
+            MVMSpeshResumeInit *resume_init = &(g->resume_inits[((MVMuint16 *)args)[1]]);
+            MVMOpInfo *res_op_info = MVM_spesh_alloc(tc, g, MVM_spesh_disp_resumption_op_info_size(
+                tc, resume_init->dp, resume_init->res_idx));
+            MVM_spesh_disp_initialize_resumption_op_info(tc,
+                    resume_init->dp, resume_init->res_idx, res_op_info);
+            info = ins_node->info = res_op_info;
+        }
+
         /* Go over operands. */
         ins_node->operands = MVM_spesh_alloc(tc, g, info->num_operands * sizeof(MVMSpeshOperand));
         for (i = 0; i < info->num_operands; i++) {
@@ -280,6 +363,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                     ins_node->operands[i].lit_i16 = GET_I16(args, arg_size);
                     arg_size += 2;
                     break;
+                case MVM_operand_uint16:
+                    ins_node->operands[i].lit_ui16 = GET_UI16(args, arg_size);
+                    arg_size += 2;
+                    break;
                 case MVM_operand_int32:
                     ins_node->operands[i].lit_i32 = GET_I32(args, arg_size);
                     arg_size += 4;
@@ -292,6 +379,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                     ins_node->operands[i].lit_i64 = MVM_BC_get_I64(args, arg_size);
                     arg_size += 8;
                     break;
+                case MVM_operand_uint64:
+                    ins_node->operands[i].lit_i64 = GET_UI64(args, arg_size);
+                    arg_size += 8;
+                    break;
                 case MVM_operand_num32:
                     ins_node->operands[i].lit_n32 = GET_N32(args, arg_size);
                     arg_size += 4;
@@ -363,24 +454,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
         }
 
-        /* Invoke and return end a basic block. Anything that is marked as
+        /* Dispatch and return end a basic block. Anything that is marked as
          * invokish and throwish are also basic block ends. OSR points are
          * basic block starts. */
-        switch (opcode) {
-        case MVM_OP_invoke_v:
-        case MVM_OP_invoke_i:
-        case MVM_OP_invoke_n:
-        case MVM_OP_invoke_s:
-        case MVM_OP_invoke_o:
-        case MVM_OP_return_i:
-        case MVM_OP_return_n:
-        case MVM_OP_return_s:
-        case MVM_OP_return_o:
-        case MVM_OP_return:
-            byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
-            next_bbs = 1;
-            break;
-        case MVM_OP_osrpoint:
+        if (opcode == MVM_OP_osrpoint) {
             byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_START;
             if (pc - g->bytecode > 0) {
                 MVMuint32 prev = pc - g->bytecode;
@@ -388,23 +465,22 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                 byte_to_ins_flags[prev] |= MVM_CFG_BB_END;
             }
             num_osr_points++;
-            break;
-        default:
-            if (info->jittivity & (MVM_JIT_INFO_THROWISH | MVM_JIT_INFO_INVOKISH)) {
-                byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
-                next_bbs = 1;
-            }
-            break;
+        }
+        else if (MVM_spesh_graph_ins_ends_bb(tc, info)) {
+            byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
+            next_bbs = 1;
         }
 
         /* Final instruction is basic block end. */
         if (pc + 2 + arg_size == end)
             byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
 
-        /* If the instruction is logged, store its program counter so we can
-         * associate it with a static value later. */
+        /* If the instruction is logged or uses the inline cache, store its
+         * program counter so we can look up the cache state or log info. */
         if (info->logged)
             add_logged_annotation(tc, g, ins_node, pc);
+        if (info->uses_cache)
+            add_cached_annotation(tc, g, ins_node, pc);
 
         /* Caculate next instruction's PC. */
         pc += 2 + arg_size;
@@ -420,6 +496,12 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             MVM_spesh_graph_add_deopt_annotation(tc, g, ins_node,
                 pc - g->bytecode, MVM_SPESH_ANN_DEOPT_OSR);
 
+        /* If it is post-logged, annotate it with the updated offset (the -2
+         * is a bit of a hack; it avoids conflicts with the following
+         * instructions pre-logged offsets). */
+        if (info->post_logged)
+            add_logged_annotation(tc, g, ins_node, pc - 2);
+
         /* Go to next instruction. */
         ins_idx++;
     }
@@ -488,7 +570,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
      * control exceptions, we will insert  */
     g->entry                  = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshBB));
     g->entry->first_ins       = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
-    g->entry->first_ins->info = get_op_info(tc, cu, 0);
+    g->entry->first_ins->info = MVM_bytecode_get_validated_op_info(tc, cu, 0);
     g->entry->last_ins        = g->entry->first_ins;
     g->entry->idx             = 0;
     cur_bb                    = NULL;
@@ -683,11 +765,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                 num_active_handlers
                 && (
                     cur_bb->last_ins->info->jittivity & (MVM_JIT_INFO_THROWISH | MVM_JIT_INFO_INVOKISH)
-                    || cur_bb->last_ins->info->opcode == MVM_OP_invoke_v
-                    || cur_bb->last_ins->info->opcode == MVM_OP_invoke_i
-                    || cur_bb->last_ins->info->opcode == MVM_OP_invoke_n
-                    || cur_bb->last_ins->info->opcode == MVM_OP_invoke_s
-                    || cur_bb->last_ins->info->opcode == MVM_OP_invoke_o
+                    || MVM_op_get_mark(cur_bb->last_ins->info->opcode)[1] == 'd'
+                    || (MVM_op_get_mark(cur_bb->last_ins->info->opcode)[1] == 's' && spesh_dispatchy(cur_bb->last_ins->info->opcode))
                 )
             ) {
                 cur_bb->handler_succ = MVM_spesh_alloc(tc, g, num_active_handlers * sizeof(MVMSpeshBB *));
@@ -716,13 +795,23 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
     if (existing_deopts) {
         for (i = 0; i < num_existing_deopts; i ++) {
             if (existing_deopts[2 * i + 1] >= 0) {
-                MVMSpeshIns *post_ins     = ins_flat[byte_to_ins_flags[existing_deopts[2 * i + 1]] >> 3];
-                MVMSpeshIns *deopt_ins    = post_ins->prev ? post_ins->prev :
-                    MVM_spesh_graph_linear_prev(tc, g,
-                        ins_to_bb[byte_to_ins_flags[existing_deopts[2 * i + 1]] >> 3])->last_ins;
+                // check last bit which is abused as PRE_INS marker and set deopt_ins = post_ins if it is set
+                MVMuint32 deopt = (MVMuint32)existing_deopts[2 * i + 1];
+                MVMuint32 bytecode_pos = MVM_spesh_deopt_bytecode_pos(deopt);
+                MVMSpeshIns *post_ins     = ins_flat[byte_to_ins_flags[bytecode_pos] >> 3];
+                MVMSpeshIns *deopt_ins;
                 MVMSpeshAnn *deopt_ann    = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
+                if (deopt & 1) { /* MVM_SPESH_ANN_DEOPT_PRE_INS annotation */
+                    deopt_ins = post_ins;
+                    deopt_ann->type = MVM_SPESH_ANN_DEOPT_PRE_INS;
+                }
+                else {
+                    deopt_ins    = post_ins->prev ? post_ins->prev :
+                        MVM_spesh_graph_linear_prev(tc, g,
+                            ins_to_bb[byte_to_ins_flags[bytecode_pos] >> 3])->last_ins;
+                    deopt_ann->type           = MVM_SPESH_ANN_DEOPT_INLINE;
+                }
                 deopt_ann->next           = deopt_ins->annotations;
-                deopt_ann->type           = MVM_SPESH_ANN_DEOPT_INLINE;
                 deopt_ann->data.deopt_idx = i;
                 deopt_ins->annotations    = deopt_ann;
             }
@@ -778,10 +867,14 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
 static MVMint32 is_handler_reg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 reg) {
     MVMuint32 num_handlers = g->num_handlers;
     MVMuint32 i;
-    for (i = 0; i < num_handlers; i++)
+    for (i = 0; i < num_handlers; i++) {
         if (g->handlers[i].action == MVM_EX_ACTION_INVOKE)
             if (g->handlers[i].block_reg == reg)
                 return 1;
+        if (g->handlers[i].category_mask & MVM_EX_CAT_LABELED)
+            if (g->handlers[i].label_reg == reg)
+                return 1;
+    }
     return 0;
 }
 static void insert_object_null_instructions(MVMThreadContext *tc, MVMSpeshGraph *g) {
@@ -1069,7 +1162,7 @@ static SSAVarInfo * initialize_ssa_var_info(MVMThreadContext *tc, MVMSpeshGraph
     return var_info;
 }
 
-MVMOpInfo *get_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint32 nrargs) {
+MVMOpInfo *MVM_spesh_graph_get_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint32 nrargs) {
     MVMOpInfo *result = NULL;
 
     /* Check number of args to phi isn't huge. */
@@ -1078,8 +1171,8 @@ MVMOpInfo *get_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint32 nrargs) {
 
     /* Up to 64 args, almost every number is represented, but after that
      * we have a sparse array through which we must search. */
-    if (nrargs - 2 < MVMPhiNodeCacheSparseBegin) {
-        result = &g->phi_infos[nrargs - 2];
+    if (nrargs - 1 < MVMPhiNodeCacheSparseBegin) {
+        result = &g->phi_infos[nrargs - 1];
     } else {
         MVMint32 cache_idx;
 
@@ -1109,9 +1202,9 @@ MVMOpInfo *get_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint32 nrargs) {
 }
 
 /* Inserts SSA phi functions at the required places in the graph. */
-static void place_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMint32 n, MVMuint16 var) {
+void MVM_spesh_graph_place_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMint32 n, MVMuint16 var) {
     MVMint32     i;
-    MVMOpInfo   *phi_op  = get_phi(tc, g, n + 1);
+    MVMOpInfo   *phi_op  = MVM_spesh_graph_get_phi(tc, g, n + 1);
     MVMSpeshIns *ins     = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
     ins->info            = phi_op;
     ins->operands        = MVM_spesh_alloc(tc, g, phi_op->num_operands * sizeof(MVMSpeshOperand));
@@ -1148,7 +1241,7 @@ static void insert_phi_functions(MVMThreadContext *tc, MVMSpeshGraph *g, SSAVarI
                 MVMSpeshBB *y = x->df[i];
                 if (has_already[y->idx] < iter_count) {
                     /* Place phi function, and mark we have. */
-                    place_phi(tc, g, y, y->num_pred, var);
+                    MVM_spesh_graph_place_phi(tc, g, y, y->num_pred, var);
                     has_already[y->idx] = iter_count;
 
                     /* Add this block to worklist if needed. */
@@ -1316,7 +1409,7 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
     g->phi_infos     = MVM_spesh_alloc(tc, g, MVMPhiNodeCacheSize * sizeof(MVMOpInfo));
 
     /* Ensure the frame is validated, since we'll rely on this. */
-    if (sf->body.instrumentation_level == 0) {
+    if (!sf->body.validated) {
         MVM_spesh_graph_destroy(tc, g);
         MVM_oops(tc, "Spesh: cannot build CFG from unvalidated frame");
     }
@@ -1350,6 +1443,9 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     g->num_lexicals      = cand->body.num_lexicals;
     g->inlines           = cand->body.inlines;
     g->num_inlines       = cand->body.num_inlines;
+    g->resume_inits      = cand->body.resume_inits;
+    g->resume_inits_num  = cand->body.num_resume_inits;
+    g->resume_inits_alloc = cand->body.num_resume_inits;
     g->deopt_addrs       = cand->body.deopts;
     g->num_deopt_addrs   = cand->body.num_deopts;
     g->alloc_deopt_addrs = cand->body.num_deopts;
@@ -1364,10 +1460,11 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
 
     g->spesh_slots       = MVM_malloc(g->alloc_spesh_slots * sizeof(MVMCollectable *));
 
-    memcpy(g->spesh_slots, cand->body.spesh_slots, sizeof(MVMCollectable *) * g->num_spesh_slots);
+    if (cand->body.spesh_slots != NULL)
+        memcpy(g->spesh_slots, cand->body.spesh_slots, sizeof(MVMCollectable *) * g->num_spesh_slots);
 
     /* Ensure the frame is validated, since we'll rely on this. */
-    if (sf->body.instrumentation_level == 0) {
+    if (!sf->body.validated) {
         MVM_spesh_graph_destroy(tc, g);
         MVM_oops(tc, "Spesh: cannot build CFG from unvalidated frame");
     }
@@ -1444,8 +1541,8 @@ void MVM_spesh_graph_mark(MVMThreadContext *tc, MVMSpeshGraph *g, MVMGCWorklist
         MVM_gc_worklist_add(tc, worklist, &(g->spesh_slots[i]));
 
     /* Mark inlines. */
-    for (i = 0; i < g->num_inlines; i++)
-        MVM_gc_worklist_add(tc, worklist, &(g->inlines[i].sf));
+    for (MVMuint32 k = 0; k < g->num_inlines; k++)
+        MVM_gc_worklist_add(tc, worklist, &(g->inlines[k].sf));
 
     /* Mark spesh candidate. */
     if (g->cand)
@@ -1484,8 +1581,8 @@ void MVM_spesh_graph_describe(MVMThreadContext *tc, MVMSpeshGraph *g, MVMHeapSna
         MVM_profile_heap_add_collectable_rel_idx(tc, snapshot, g->spesh_slots[i], i);
 
     /* Mark inlines. */
-    for (i = 0; i < g->num_inlines; i++)
-        MVM_profile_heap_add_collectable_rel_idx(tc, snapshot, (MVMCollectable *)g->inlines[i].sf, i);
+    for (MVMuint32 k = 0; k < g->num_inlines; k++)
+        MVM_profile_heap_add_collectable_rel_idx(tc, snapshot, (MVMCollectable *)g->inlines[k].sf, k);
 }
 
 /* Destroys a spesh graph, deallocating all its associated memory. */
@@ -1501,6 +1598,8 @@ void MVM_spesh_graph_destroy(MVMThreadContext *tc, MVMSpeshGraph *g) {
         MVM_free(g->deopt_addrs);
     if (g->inlines && (!g->cand || g->cand->body.inlines != g->inlines))
         MVM_free(g->inlines);
+    if (g->resume_inits && (!g->cand || g->cand->body.resume_inits != g->resume_inits))
+        MVM_free(g->resume_inits);
     if (g->local_types &&  (!g->cand || g->cand->body.local_types != g->local_types))
         MVM_free(g->local_types);
     if (g->lexical_types &&  (!g->cand || g->cand->body.lexical_types != g->lexical_types))
