@@ -1,7 +1,7 @@
 #include "moar.h"
 
 /* Gets the statistics for a static frame, creating them if needed. */
-MVMSpeshStats * stats_for(MVMThreadContext *tc, MVMStaticFrame *sf) {
+static MVMSpeshStats * stats_for(MVMThreadContext *tc, MVMStaticFrame *sf) {
     MVMStaticFrameSpesh *spesh = sf->body.spesh;
     if (!spesh->body.spesh_stats)
         spesh->body.spesh_stats = MVM_calloc(1, sizeof(MVMSpeshStats));
@@ -9,7 +9,7 @@ MVMSpeshStats * stats_for(MVMThreadContext *tc, MVMStaticFrame *sf) {
 }
 
 /* Gets the stats by callsite, adding it if it's missing. */
-MVMuint32 by_callsite_idx(MVMThreadContext *tc, MVMSpeshStats *ss, MVMCallsite *cs) {
+static MVMuint32 by_callsite_idx(MVMThreadContext *tc, MVMSpeshStats *ss, MVMCallsite *cs) {
     /* See if we already have it. */
     MVMuint32 found;
     MVMuint32 n = ss->num_by_callsite;
@@ -29,8 +29,8 @@ MVMuint32 by_callsite_idx(MVMThreadContext *tc, MVMSpeshStats *ss, MVMCallsite *
 
 /* Checks if a type tuple is incomplete (no types logged for some passed
  * objects, or no decont type logged for a container type). */
-MVMint32 incomplete_type_tuple(MVMThreadContext *tc, MVMCallsite *cs,
-                               MVMSpeshStatsType *arg_types) {
+static MVMint32 incomplete_type_tuple(MVMThreadContext *tc, MVMCallsite *cs,
+                                      MVMSpeshStatsType *arg_types) {
     MVMuint32 i;
     for (i = 0; i < cs->flag_count; i++) {
         if (cs->arg_flags[i] & MVM_CALLSITE_ARG_OBJ) {
@@ -46,7 +46,7 @@ MVMint32 incomplete_type_tuple(MVMThreadContext *tc, MVMCallsite *cs,
 }
 
 /* Returns true if the callsite has no object arguments, false otherwise. */
-MVMint32 cs_without_object_args(MVMThreadContext *tc, MVMCallsite *cs) {
+static MVMint32 cs_without_object_args(MVMThreadContext *tc, MVMCallsite *cs) {
     MVMuint32 i;
     for (i = 0; i < cs->flag_count; i++)
         if (cs->arg_flags[i] & MVM_CALLSITE_ARG_OBJ)
@@ -56,8 +56,8 @@ MVMint32 cs_without_object_args(MVMThreadContext *tc, MVMCallsite *cs) {
 
 /* Gets the stats by type, adding it if it's missing. Frees arg_types. Returns
  * the index in the by type array, or -1 if unresolved. */
-MVMint32 by_type(MVMThreadContext *tc, MVMSpeshStats *ss, MVMuint32 callsite_idx,
-                  MVMSpeshStatsType *arg_types) {
+static MVMint32 by_type(MVMThreadContext *tc, MVMSpeshStats *ss, MVMuint32 callsite_idx,
+                        MVMSpeshStatsType *arg_types) {
     /* Resolve type by callsite level info. If this is the no-callsite
      * specialization there is nothing further to do. */
     MVMSpeshStatsByCallsite *css = &(ss->by_callsite[callsite_idx]);
@@ -113,8 +113,8 @@ MVMint32 by_type(MVMThreadContext *tc, MVMSpeshStats *ss, MVMuint32 callsite_idx
 }
 
 /* Get the stats by offset entry, adding it if it's missing. */
-MVMSpeshStatsByOffset * by_offset(MVMThreadContext *tc, MVMSpeshStatsByType *tss,
-                                  MVMuint32 bytecode_offset) {
+static MVMSpeshStatsByOffset * by_offset(MVMThreadContext *tc, MVMSpeshStatsByType *tss,
+                                         MVMuint32 bytecode_offset) {
     /* See if we already have it. */
     MVMuint32 found;
     MVMuint32 n = tss->num_by_offset;
@@ -133,8 +133,8 @@ MVMSpeshStatsByOffset * by_offset(MVMThreadContext *tc, MVMSpeshStatsByType *tss
 }
 
 /* Adds/increments the count of a certain type seen at the given offset. */
-void add_type_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
-                        MVMStaticFrame *sf, MVMObject *type, MVMuint8 concrete) {
+static void add_type_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
+                               MVMStaticFrame *sf, MVMObject *type, MVMuint8 concrete) {
     /* If we have it already, increment the count. */
     MVMuint32 found;
     MVMuint32 n = oss->num_types;
@@ -156,9 +156,9 @@ void add_type_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
 
 /* Adds/increments the count of a certain invocation target seen at the given
  * offset. */
-void add_invoke_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
-                          MVMStaticFrame *sf, MVMStaticFrame *target_sf,
-                          MVMint16 caller_is_outer, MVMint16 was_multi) {
+static void add_invoke_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
+                                 MVMStaticFrame *sf, MVMStaticFrame *target_sf,
+                                 MVMint16 caller_is_outer) {
     /* If we have it already, increment the count. */
     MVMuint32 found;
     MVMuint32 n = oss->num_invokes;
@@ -167,8 +167,6 @@ void add_invoke_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
             oss->invokes[found].count++;
             if (caller_is_outer)
                 oss->invokes[found].caller_is_outer_count++;
-            if (was_multi)
-                oss->invokes[found].was_multi_count++;
             return;
         }
     }
@@ -181,34 +179,33 @@ void add_invoke_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
     MVM_ASSIGN_REF(tc, &(sf->body.spesh->common.header), oss->invokes[found].sf, target_sf);
     oss->invokes[found].count = 1;
     oss->invokes[found].caller_is_outer_count = caller_is_outer ? 1 : 0;
-    oss->invokes[found].was_multi_count = was_multi ? 1 : 0;
 }
 
-/* Adds/increments the count of a plugin guard index seen at the given offset. */
-void add_plugin_guard_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
-                                MVMuint32 guard_index) {
+/* Adds/increments the count of a dispatch result seen at the given offset. */
+static void add_dispatch_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
+                                   MVMuint32 result_index) {
     /* If we have it already, increment the count. */
     MVMuint32 found;
-    MVMuint32 n = oss->num_plugin_guards;
+    MVMuint32 n = oss->num_dispatch_results;
     for (found = 0; found < n; found++) {
-        if (oss->plugin_guards[found].guard_index == guard_index) {
-            oss->plugin_guards[found].count++;
+        if (oss->dispatch_results[found].result_index == result_index) {
+            oss->dispatch_results[found].count++;
             return;
         }
     }
 
     /* Otherwise, add it to the list. */
-    found = oss->num_plugin_guards;
-    oss->num_plugin_guards++;
-    oss->plugin_guards = MVM_realloc(oss->plugin_guards,
-            oss->num_plugin_guards * sizeof(MVMSpeshStatsPluginGuardCount));
-    oss->plugin_guards[found].guard_index = guard_index;
-    oss->plugin_guards[found].count = 1;
+    found = oss->num_dispatch_results;
+    oss->num_dispatch_results++;
+    oss->dispatch_results = MVM_realloc(oss->dispatch_results,
+            oss->num_dispatch_results * sizeof(MVMSpeshStatsDispatchResultCount));
+    oss->dispatch_results[found].result_index = result_index;
+    oss->dispatch_results[found].count = 1;
 }
 
 /* Adds/increments the count of a type tuple seen at the given offset. */
-void add_type_tuple_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
-                              MVMStaticFrame *sf, MVMSpeshSimCallType *info) {
+static void add_type_tuple_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
+                                     MVMStaticFrame *sf, MVMSpeshSimCallType *info) {
     /* Compute type tuple size. */
     size_t tt_size = info->cs->flag_count * sizeof(MVMSpeshStatsType);
 
@@ -245,7 +242,7 @@ void add_type_tuple_at_offset(MVMThreadContext *tc, MVMSpeshStatsByOffset *oss,
 }
 
 /* Initializes the stack simulation. */
-void sim_stack_init(MVMThreadContext *tc, MVMSpeshSimStack *sims) {
+static void sim_stack_init(MVMThreadContext *tc, MVMSpeshSimStack *sims) {
     sims->used = 0;
     sims->limit = 32;
     sims->frames = MVM_malloc(sims->limit * sizeof(MVMSpeshSimStackFrame));
@@ -253,8 +250,8 @@ void sim_stack_init(MVMThreadContext *tc, MVMSpeshSimStack *sims) {
 }
 
 /* Pushes an entry onto the stack frame model. */
-void sim_stack_push(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMStaticFrame *sf,
-                    MVMSpeshStats *ss, MVMuint32 cid, MVMuint32 callsite_idx) {
+static void sim_stack_push(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMStaticFrame *sf,
+                           MVMSpeshStats *ss, MVMuint32 cid, MVMuint32 callsite_idx) {
     MVMSpeshSimStackFrame *frame;
     MVMCallsite *cs;
     if (sims->used == sims->limit) {
@@ -282,9 +279,9 @@ void sim_stack_push(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMStaticFrame
 
 /* Adds an entry to a sim frame's callsite type info list, for later
  * inclusion in the callsite stats. */
-void add_sim_call_type_info(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
-                            MVMuint32 bytecode_offset, MVMCallsite *cs,
-                            MVMSpeshStatsType *arg_types) {
+static void add_sim_call_type_info(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
+                                   MVMuint32 bytecode_offset, MVMCallsite *cs,
+                                   MVMSpeshStatsType *arg_types) {
     MVMSpeshSimCallType *info;
     if (simf->call_type_info_used == simf->call_type_info_limit) {
         simf->call_type_info_limit += 32;
@@ -298,9 +295,9 @@ void add_sim_call_type_info(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
 }
 
 /* Incorporate information collected into a sim stack frame. */
-void incorporate_stats(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
-                       MVMuint32 frame_depth, MVMSpeshSimStackFrame *caller,
-                       MVMObject *sf_updated) {
+static void incorporate_stats(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
+                              MVMuint32 frame_depth, MVMSpeshSimStackFrame *caller,
+                              MVMObject *sf_updated) {
     MVMSpeshStatsByType *tss;
     MVMint32 first_type_hit = 0;
 
@@ -345,13 +342,13 @@ void incorporate_stats(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
                     MVMSpeshStatsByOffset *oss = by_offset(tc, tss,
                         e->invoke.bytecode_offset);
                     add_invoke_at_offset(tc, oss, simf->sf, e->invoke.sf,
-                        e->invoke.caller_is_outer, e->invoke.was_multi);
+                        e->invoke.caller_is_outer);
                     break;
                 }
-                case MVM_SPESH_LOG_PLUGIN_RESOLUTION: {
+                case MVM_SPESH_LOG_DISPATCH_RESOLUTION: {
                     MVMSpeshStatsByOffset *oss = by_offset(tc, tss,
-                        e->plugin.bytecode_offset);
-                    add_plugin_guard_at_offset(tc, oss, e->plugin.guard_index);
+                        e->dispatch.bytecode_offset);
+                    add_dispatch_at_offset(tc, oss, e->dispatch.result_index);
                     break;
                 }
             }
@@ -391,7 +388,7 @@ void incorporate_stats(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf,
 }
 
 /* Pops the top frame from the sim stack. */
-void sim_stack_pop(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMObject *sf_updated) {
+static void sim_stack_pop(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMObject *sf_updated) {
     MVMSpeshSimStackFrame *simf;
     MVMuint32 frame_depth;
 
@@ -413,8 +410,8 @@ void sim_stack_pop(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMObject *sf_u
  * not on the top, searches to see if it's further down. If it is, then pops
  * off the top to reach it. If it's not found at all, returns NULL and does
  * nothing to the simulation stack. */
-MVMSpeshSimStackFrame * sim_stack_find(MVMThreadContext *tc, MVMSpeshSimStack *sims,
-                                       MVMuint32 cid, MVMObject *sf_updated) {
+static MVMSpeshSimStackFrame * sim_stack_find(MVMThreadContext *tc, MVMSpeshSimStack *sims,
+                                              MVMuint32 cid, MVMObject *sf_updated) {
     MVMuint32 found_at = sims->used;
     while (found_at != 0) {
         found_at--;
@@ -430,45 +427,27 @@ MVMSpeshSimStackFrame * sim_stack_find(MVMThreadContext *tc, MVMSpeshSimStack *s
 }
 
 /* Pops all the frames in the stack simulation and frees the frames storage. */
-void sim_stack_teardown(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMObject *sf_updated) {
+static void sim_stack_teardown(MVMThreadContext *tc, MVMSpeshSimStack *sims, MVMObject *sf_updated) {
     while (sims->used)
         sim_stack_pop(tc, sims, sf_updated);
     MVM_free(sims->frames);
 }
 
 /* Gets the parameter type slot from a simulation frame. */
-MVMSpeshStatsType * param_type(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf, MVMSpeshLogEntry *e) {
+static MVMSpeshStatsType * param_type(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf, MVMSpeshLogEntry *e) {
     if (simf->arg_types) {
         MVMuint16 idx = e->param.arg_idx;
         MVMCallsite *cs = simf->ss->by_callsite[simf->callsite_idx].cs;
         if (cs) {
-            MVMint32 flag_idx = idx < cs->num_pos
-                ? idx
-                : cs->num_pos + (((idx - 1) - cs->num_pos) / 2);
-            if (flag_idx >= cs->flag_count)
+            if (idx >= cs->flag_count)
                 MVM_panic(1, "Spesh stats: argument flag index out of bounds");
-            if (cs->arg_flags[flag_idx] & MVM_CALLSITE_ARG_OBJ)
-                return &(simf->arg_types[flag_idx]);
+            if (cs->arg_flags[idx] & MVM_CALLSITE_ARG_OBJ)
+                return &(simf->arg_types[idx]);
         }
     }
     return NULL;
 }
 
-/* Records a static value for a frame, unless it's already in the log. */
-void add_static_value(MVMThreadContext *tc, MVMSpeshSimStackFrame *simf, MVMuint32 bytecode_offset,
-                      MVMObject *value) {
-    MVMSpeshStats *ss = simf->ss;
-    MVMuint32 i, id;
-    for (i = 0; i < ss->num_static_values; i++)
-        if (ss->static_values[i].bytecode_offset == bytecode_offset)
-            return;
-    id = ss->num_static_values++;
-    ss->static_values = MVM_realloc(ss->static_values,
-        ss->num_static_values * sizeof(MVMSpeshStatsStatic));
-    ss->static_values[id].bytecode_offset = bytecode_offset;
-    MVM_ASSIGN_REF(tc, &(simf->sf->body.spesh->common.header), ss->static_values[id].value, value);
-}
-
 /* Decides whether to save or free the simulation stack. */
 static void save_or_free_sim_stack(MVMThreadContext *tc, MVMSpeshSimStack *sims,
                                    MVMThreadContext *save_on_tc, MVMObject *sf_updated) {
@@ -519,7 +498,8 @@ static void save_or_free_sim_stack(MVMThreadContext *tc, MVMSpeshSimStack *sims,
 
 /* Receives a spesh log and updates static frame statistics. Each static frame
  * that is updated is pushed once into sf_updated. */
-void MVM_spesh_stats_update(MVMThreadContext *tc, MVMSpeshLog *sl, MVMObject *sf_updated, MVMuint64 *in_newly_seen, MVMuint64 *in_updated) {
+void MVM_spesh_stats_update(MVMThreadContext *tc, MVMSpeshLog *sl,  MVMObject *sf_newly_seen,
+        MVMObject *sf_updated, MVMuint64 *in_newly_seen, MVMuint64 *in_updated) {
     MVMuint32 i;
     MVMuint32 n = sl->body.used;
     MVMSpeshSimStack *sims;
@@ -558,10 +538,13 @@ void MVM_spesh_stats_update(MVMThreadContext *tc, MVMSpeshLog *sl, MVMObject *sf
             case MVM_SPESH_LOG_ENTRY: {
                 MVMSpeshStats *ss = stats_for(tc, e->entry.sf);
                 MVMuint32 callsite_idx;
-                if (ss->last_update == 0)
+                if (ss->last_update == 0) {
                     newly_seen++;
-                else
+                    MVM_repr_push_o(tc, sf_newly_seen, (MVMObject *)e->entry.sf);
+                }
+                else {
                     updated++;
+                }
                 if (ss->last_update != tc->instance->spesh_stats_version) {
                     ss->last_update = tc->instance->spesh_stats_version;
                     MVM_repr_push_o(tc, sf_updated, (MVMObject *)e->entry.sf);
@@ -603,7 +586,7 @@ void MVM_spesh_stats_update(MVMThreadContext *tc, MVMSpeshLog *sl, MVMObject *sf
             case MVM_SPESH_LOG_TYPE:
             case MVM_SPESH_LOG_RETURN:
             case MVM_SPESH_LOG_INVOKE:
-            case MVM_SPESH_LOG_PLUGIN_RESOLUTION: {
+            case MVM_SPESH_LOG_DISPATCH_RESOLUTION: {
                 /* We only incorporate these into the model later, and only
                  * then if we need to. For now, just keep references to
                  * them. */
@@ -628,12 +611,6 @@ void MVM_spesh_stats_update(MVMThreadContext *tc, MVMSpeshLog *sl, MVMObject *sf
                     simf->osr_hits++;
                 break;
             }
-            case MVM_SPESH_LOG_STATIC: {
-                MVMSpeshSimStackFrame *simf = sim_stack_find(tc, sims, e->id, sf_updated);
-                if (simf)
-                    add_static_value(tc, simf, e->value.bytecode_offset, e->value.value);
-                break;
-            }
             case MVM_SPESH_LOG_RETURN_TO_UNLOGGED: {
                 MVMSpeshSimStackFrame *simf = sim_stack_find(tc, sims, e->id, sf_updated);
                 if (simf)
@@ -657,17 +634,23 @@ void MVM_spesh_stats_update(MVMThreadContext *tc, MVMSpeshLog *sl, MVMObject *sf
  * updated in a while, clears them out. */
 void MVM_spesh_stats_cleanup(MVMThreadContext *tc, MVMObject *check_frames) {
     MVMint64 elems = MVM_repr_elems(tc, check_frames);
-    MVMROOT(tc, check_frames, {
+    MVMSTable *check_frames_st = STABLE(check_frames);
+    void *check_frames_data = OBJECT_BODY(check_frames);
+    MVMROOT(tc, check_frames) {
         MVMint64 insert_pos = 0;
         MVMint64 i;
         for (i = 0; i < elems; i++) {
-            MVMStaticFrame *sf = (MVMStaticFrame *)MVM_repr_at_pos_o(tc, check_frames, i);
-            MVMROOT(tc, sf, {
+            MVMRegister sf_reg;
+            MVM_VMArray_at_pos_o(tc, check_frames_st, check_frames, check_frames_data, i, &sf_reg);
+            MVMStaticFrame *sf = (MVMStaticFrame *)sf_reg.o;
+            MVMROOT(tc, sf) {
                 MVMStaticFrameSpesh *spesh = sf->body.spesh;
                 MVMSpeshStats *ss = spesh->body.spesh_stats;
+                MVMuint32 removed = 0;
                 if (!ss) {
                     /* No stats; already destroyed, don't keep this frame under
                      * consideration. */
+                    removed = 1;
                 }
                 else if (tc->instance->spesh_stats_version - ss->last_update > MVM_SPESH_STATS_MAX_AGE) {
                     /* Do not mark thread blocked as the GC also tries to acquire
@@ -698,18 +681,19 @@ void MVM_spesh_stats_cleanup(MVMThreadContext *tc, MVMObject *check_frames) {
                     if (!found) {
                         MVM_spesh_stats_destroy(tc, ss);
                         MVM_free_null(spesh->body.spesh_stats);
-                    }
-                    else {
-                        MVM_repr_bind_pos_o(tc, check_frames, insert_pos++, (MVMObject *)sf);
+                        removed = 1;
                     }
                 }
-                else {
-                    MVM_repr_bind_pos_o(tc, check_frames, insert_pos++, (MVMObject *)sf);
+
+                if (!removed) {
+                    sf_reg.o = (MVMObject *)sf;
+                    MVM_VMArray_bind_pos(tc, check_frames_st, check_frames,
+                            check_frames_data, insert_pos++, sf_reg, MVM_reg_obj);
                 }
-            });
+            }
         }
         MVM_repr_pos_set_elems(tc, check_frames, insert_pos);
-    });
+    }
 }
 
 void MVM_spesh_stats_gc_mark(MVMThreadContext *tc, MVMSpeshStats *ss, MVMGCWorklist *worklist) {
@@ -741,8 +725,6 @@ void MVM_spesh_stats_gc_mark(MVMThreadContext *tc, MVMSpeshStats *ss, MVMGCWorkl
                 }
             }
         }
-        for (i = 0; i < ss->num_static_values; i++)
-            MVM_gc_worklist_add(tc, worklist, &(ss->static_values[i].value));
     }
 }
 
@@ -752,7 +734,6 @@ void MVM_spesh_stats_gc_describe(MVMThreadContext *tc, MVMHeapSnapshotState *sna
     MVMuint64 cache_3 = 0;
     MVMuint64 cache_4 = 0;
     MVMuint64 cache_5 = 0;
-    MVMuint64 cache_6 = 0;
     if (ss) {
         MVMuint32 i, j, k, l, m;
         for (i = 0; i < ss->num_by_callsite; i++) {
@@ -787,9 +768,6 @@ void MVM_spesh_stats_gc_describe(MVMThreadContext *tc, MVMHeapSnapshotState *sna
                 }
             }
         }
-        for (i = 0; i < ss->num_static_values; i++)
-            MVM_profile_heap_add_collectable_rel_const_cstr_cached(tc, snapshot,
-                (MVMCollectable*)(ss->static_values[i].value), "static value", &cache_6);
     }
 }
 
@@ -807,7 +785,7 @@ void MVM_spesh_stats_destroy(MVMThreadContext *tc, MVMSpeshStats *ss) {
                     for (l = 0; l < by_offset->num_type_tuples; l++)
                         MVM_free(by_offset->type_tuples[l].arg_types);
                     MVM_free(by_offset->type_tuples);
-                    MVM_free(by_offset->plugin_guards);
+                    MVM_free(by_offset->dispatch_results);
                 }
                 MVM_free(by_type->by_offset);
                 MVM_free(by_type->arg_types);
@@ -815,7 +793,6 @@ void MVM_spesh_stats_destroy(MVMThreadContext *tc, MVMSpeshStats *ss) {
             MVM_free(by_cs->by_type);
         }
         MVM_free(ss->by_callsite);
-        MVM_free(ss->static_values);
     }
 }
 
