@@ -174,6 +174,36 @@ static void reached_eof(MVMThreadContext *tc, MVMDecodeStream *ds) {
     /* Decode all the things. */
     if (ds->bytes_head)
         run_decode(tc, ds, NULL, NULL, DECODE_EOF);
+    if (ds->bytes_head){
+        MVMDecodeStreamBytes *bh = ds->bytes_head;
+        MVMint32 i = ds->bytes_head_pos;
+        char dumped[16] = " xx xx xx xx...";
+        size_t j = 0;
+        while(bh && j < sizeof(dumped) - 4){
+            if(i < bh->length){
+                int r = snprintf(&dumped[j], sizeof(dumped)-j,
+                    " %02hhx", bh->bytes[i++]);
+                if(r < 0){
+                    j = 0;
+                    break;
+                }
+                j += r;
+            }
+            if(i >= bh->length){
+                bh = bh->next;
+                i = 0;
+            }
+        }
+        if(j == 0){
+            MVM_exception_throw_adhoc(tc, "Incomplete character "
+                "at the end of a stream");
+        }else{
+            if(bh)
+                dumped[12] = '.';
+            MVM_exception_throw_adhoc(tc, "Incomplete character "
+                "near bytes%s at the end of a stream", dumped);
+        }
+    }
 
     /* If there's some things left in the normalization buffer, take them. */
     MVM_unicode_normalizer_eof(tc, &(ds->norm));
@@ -442,18 +472,19 @@ MVMString * MVM_string_decodestream_get_until_sep_eof(MVMThreadContext *tc, MVMD
 /* Produces a string consisting of the characters available now in all decdoed
  * buffers. */
 static MVMString * get_all_in_buffer(MVMThreadContext *tc, MVMDecodeStream *ds) {
-    MVMString *result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
-    result->body.storage_type = MVM_STRING_GRAPHEME_32;
+    MVMString *result;
 
     /* If there's no codepoint buffer, then return the empty string. */
     if (!ds->chars_head) {
-        result->body.storage.blob_32 = NULL;
-        result->body.num_graphs      = 0;
+        return tc->instance->str_consts.empty;
     }
 
+    result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
+    result->body.storage_type = MVM_STRING_GRAPHEME_32;
+
     /* If there's exactly one resulting codepoint buffer and we swallowed none
      * of it, just use it. */
-    else if (ds->chars_head == ds->chars_tail && ds->chars_head_pos == 0) {
+    if (ds->chars_head == ds->chars_tail && ds->chars_head_pos == 0) {
         /* Set up result string. */
         result->body.storage.blob_32 = ds->chars_head->chars;
         result->body.num_graphs      = ds->chars_head->length;
@@ -664,7 +695,7 @@ void MVM_string_decode_stream_sep_default(MVMThreadContext *tc, MVMDecodeStreamS
 void MVM_string_decode_stream_sep_from_strings(MVMThreadContext *tc, MVMDecodeStreamSeparators *sep_spec,
                                                MVMString **seps, MVMint32 num_seps) {
     MVMGraphemeIter gi;
-    MVMint32 i, graph_length, graph_pos;
+    MVMint32 i, graph_length, graph_pos, *sep_lengths;
 
     if (num_seps > 0xFFF)
         MVM_exception_throw_adhoc(tc, "Too many line separators (%"PRId32"), max allowed is 4095", num_seps);
@@ -674,15 +705,18 @@ void MVM_string_decode_stream_sep_from_strings(MVMThreadContext *tc, MVMDecodeSt
     MVM_free(sep_spec->final_graphemes);
 
     sep_spec->num_seps = num_seps;
-    sep_spec->sep_lengths = MVM_malloc(num_seps * sizeof(MVMint32));
+    sep_lengths = MVM_malloc(num_seps * sizeof(MVMint32));
     graph_length = 0;
     for (i = 0; i < num_seps; i++) {
         MVMuint32 num_graphs = MVM_string_graphs(tc, seps[i]);
-        if (num_graphs > 0xFFFF)
+        if (num_graphs > 0xFFFF) {
+            MVM_free(sep_lengths);
             MVM_exception_throw_adhoc(tc, "Line separator (%"PRIu32") too long, max allowed is 65535", num_graphs);
-        sep_spec->sep_lengths[i] = num_graphs;
+        }
+        sep_lengths[i] = num_graphs;
         graph_length += num_graphs;
     }
+    sep_spec->sep_lengths = sep_lengths;
 
     sep_spec->sep_graphemes = MVM_malloc(graph_length * sizeof(MVMGrapheme32));
     graph_pos = 0;
