@@ -1,46 +1,86 @@
 #include "moar.h"
 
+#define DECODE_BODY                                                          \
+    for (i = 0; i < bytes; i++) {                                            \
+        if (latin1[i] == '\r' && i + 1 < bytes && latin1[i + 1] == '\n') {   \
+            storage[result_graphs++] = MVM_nfg_crlf_grapheme(tc);            \
+            i++;                                                             \
+        }                                                                    \
+        else {                                                               \
+            storage[result_graphs++] = latin1[i];                            \
+        }                                                                    \
+    }                                                                        \
+
+#define DECODE_BODY_NO_CRLF                                                  \
+    for (i = 0; i < bytes; i++) {                                            \
+        storage[result_graphs++] = latin1[i];                                \
+    }                                                                        \
+
 /* Decodes the specified number of bytes of latin1 into an NFG string,
  * creating a result of the specified type. The type must have the MVMString
  * REPR. */
 MVMString * MVM_string_latin1_decode(MVMThreadContext *tc, const MVMObject *result_type,
                                      char *latin1_c, size_t bytes) {
     MVMuint8  *latin1 = (MVMuint8 *)latin1_c;
-    MVMString *result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
-    size_t i, k, result_graphs;
+    MVMString *result;
+    size_t i, result_graphs;
+
+    if (bytes == 0 && tc->instance->str_consts.empty) {
+        return tc->instance->str_consts.empty;
+    }
 
     MVMuint8 writing_32bit = 0;
+    MVMuint8 has_crlf = 0;
+    MVM_VECTORIZE_LOOP
+    for (i = 0; i < bytes - 1; i++) {
+        writing_32bit |= (latin1[i] > 127);
+        has_crlf |= (latin1[i] == '\r' && latin1[i + 1] == '\n');
+    }
+    writing_32bit |= (latin1[i] > 127);
+    has_crlf |= (latin1[i - 1] == '\r' && latin1[i] == '\n');
 
-    result->body.storage_type   = MVM_STRING_GRAPHEME_8;
-    result->body.storage.blob_8 = MVM_malloc(sizeof(MVMint8) * bytes);
+    result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
 
     result_graphs = 0;
-    for (i = 0; i < bytes; i++) {
-        if (latin1[i] == '\r' && i + 1 < bytes && latin1[i + 1] == '\n') {
-            if (writing_32bit)
-                result->body.storage.blob_32[result_graphs++] = MVM_nfg_crlf_grapheme(tc);
-            else
-                result->body.storage.blob_8[result_graphs++] = MVM_nfg_crlf_grapheme(tc);
-            i++;
+    if (writing_32bit) {
+        MVMGrapheme32 *storage;
+        if (bytes <= 2) {
+            result->body.storage_type    = MVM_STRING_IN_SITU_32;
+            storage = result->body.storage.in_situ_32;
         }
         else {
-            if (latin1[i] > 127 && !writing_32bit) {
-                MVMGrapheme8 *old_storage = result->body.storage.blob_8;
-
-                result->body.storage.blob_32 = MVM_malloc(sizeof(MVMGrapheme32) * bytes);
-                result->body.storage_type = MVM_STRING_GRAPHEME_32;
-                writing_32bit = 1;
-
-                for (k = 0; k < i; k++)
-                    result->body.storage.blob_32[k] = old_storage[k];
-                MVM_free(old_storage);
-            }
-            if (writing_32bit)
-                result->body.storage.blob_32[result_graphs++] = latin1[i];
-            else
-                result->body.storage.blob_8[result_graphs++] = latin1[i];
+            result->body.storage_type    = MVM_STRING_GRAPHEME_32;
+            result->body.storage.blob_32 = MVM_malloc(sizeof(MVMGrapheme32) * bytes);
+            storage = result->body.storage.blob_32;
+        }
+        if (has_crlf) {
+            DECODE_BODY
+        }
+        else {
+            MVM_VECTORIZE_LOOP
+            DECODE_BODY_NO_CRLF
         }
     }
+    else {
+        MVMint8 *storage;
+        if (bytes <= 8) {
+            result->body.storage_type   = MVM_STRING_IN_SITU_8;
+            storage = result->body.storage.in_situ_8;
+        }
+        else {
+            result->body.storage_type   = MVM_STRING_GRAPHEME_8;
+            result->body.storage.blob_8 = MVM_malloc(sizeof(MVMint8) * bytes);
+            storage = result->body.storage.blob_8;
+        }
+        if (has_crlf) {
+            DECODE_BODY
+        }
+        else {
+            MVM_VECTORIZE_LOOP
+            DECODE_BODY_NO_CRLF
+        }
+    }
+
     result->body.num_graphs = result_graphs;
 
     return result;
@@ -49,10 +89,10 @@ MVMString * MVM_string_latin1_decode(MVMThreadContext *tc, const MVMObject *resu
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
 MVMuint32 MVM_string_latin1_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
-                                    const MVMint32 *stopper_chars,
+                                    const MVMuint32 *stopper_chars,
                                     MVMDecodeStreamSeparators *seps) {
-    MVMint32 count = 0, total = 0;
-    MVMint32 bufsize;
+    MVMuint32 count = 0, total = 0;
+    MVMuint32 bufsize;
     MVMGrapheme32 *buffer;
     MVMDecodeStreamBytes *cur_bytes;
     MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
@@ -78,7 +118,7 @@ MVMuint32 MVM_string_latin1_decodestream(MVMThreadContext *tc, MVMDecodeStream *
     while (cur_bytes) {
         /* Process this buffer. */
         MVMint32  pos = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
-        unsigned char *bytes = (unsigned char *)cur_bytes->bytes;
+        MVMuint8 *bytes = cur_bytes->bytes;
         while (pos < cur_bytes->length) {
             MVMCodepoint codepoint = bytes[pos++];
             MVMGrapheme32 graph;
@@ -149,9 +189,9 @@ char * MVM_string_latin1_encode_substr(MVMThreadContext *tc, MVMString *str, MVM
 
     /* must check start first since it's used in the length check */
     if (start < 0 || start > strgraphs)
-        MVM_exception_throw_adhoc(tc, "start out of range");
+        MVM_exception_throw_adhoc(tc, "start (%"PRId64") out of range (0..%"PRIu32")", start, strgraphs);
     if (length < -1 || start + lengthu > strgraphs)
-        MVM_exception_throw_adhoc(tc, "length out of range");
+        MVM_exception_throw_adhoc(tc, "length (%"PRId64") out of range (-1..%"PRIu32")", length, strgraphs);
 
     if (replacement)
         repl_bytes = (MVMuint8 *) MVM_string_latin1_encode_substr(tc,
