@@ -204,7 +204,7 @@ MVMString * MVM_string_shiftjis_decode(MVMThreadContext *tc,
         const MVMObject *result_type, char *windows125X_c, size_t num_bytes,
         MVMString *replacement, MVMint64 config) {
     MVMuint8 *bytes = (MVMuint8 *)windows125X_c;
-    MVMString *result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
+    MVMString *result;
     size_t pos = 0, result_graphs;
     MVMStringIndex repl_length = replacement ? MVM_string_graphs(tc, replacement) : 0;
     MVMuint8 Shift_JIS_lead = 0x00;
@@ -217,30 +217,44 @@ MVMString * MVM_string_shiftjis_decode(MVMThreadContext *tc,
     int last_was_cr         = 0;
     MVMStringIndex result_size = num_bytes;
 
-    result->body.storage_type    = MVM_STRING_GRAPHEME_32;
     /* TODO allocate less? */
-    result->body.storage.blob_32 = MVM_malloc(sizeof(MVMGrapheme32) * result_size);
+    MVMGrapheme32 *buffer = MVM_malloc(sizeof(MVMGrapheme32) * result_size);
 
     result_graphs = 0;
-    while (pos < num_bytes || repl_pos) {
+    int start_repl = 0;
+    while (pos < num_bytes || start_repl) {
         MVMGrapheme32 graph;
         MVMGrapheme32 codepoint;
         MVMuint8 byte;
-        int graph_is_set = 0;
-        if (repl_pos) {
+        if (start_repl) {
+            /* Except for the case of the previous grapheme being \r and the replacement
+             * starting with \n, we shouldn't get any normalization or codepoints combining
+             * to make graphemes */
             graph = MVM_string_get_grapheme_at_nocheck(tc, replacement, repl_pos++);
-            graph_is_set = 1;
-            if (repl_length <= repl_pos) repl_pos = 0;
-        }
-        else if (is_prepended) {
-            byte = prepended;
-            is_prepended = 0;
+            if (last_was_cr) {
+                if (graph == '\n') {
+                    graph = MVM_nfg_crlf_grapheme(tc);
+                }
+                else {
+                    graph = '\r';
+                    repl_pos--;
+                }
+                last_was_cr = 0;
+            }
+            if (repl_length <= repl_pos) {
+                repl_pos   = 0;
+                start_repl = 0;
+            }
         }
+        /* Main path. There was no replacement */
         else {
-            byte = bytes[pos++];
-        }
-        /* graph_is_set will be 0 unless we just grabbed a replacement grapheme */
-        if (!graph_is_set) {
+            if (is_prepended) {
+                byte = prepended;
+                is_prepended = 0;
+            }
+            else {
+                byte = bytes[pos++];
+            }
             int handler_rtrn = decoder_handler(tc, &Shift_JIS_lead, byte, &codepoint);
             if (handler_rtrn == DECODE_CODEPOINT) {
                 graph = codepoint;
@@ -253,13 +267,11 @@ MVMString * MVM_string_shiftjis_decode(MVMThreadContext *tc,
                  * a replacement is used. */
                 Shift_JIS_lead = 0x00;
                 if (replacement) {
-                    graph = MVM_string_get_grapheme_at_nocheck(tc, replacement, repl_pos);
-                    /* If the replacement is more than one grapheme we need
-                     * to set repl_pos++ so we will grab the next grapheme on
-                     * the next loop */
-                    if (1 < repl_length) repl_pos++;
+                    start_repl = 1;
+                    continue;
                 }
                 else {
+                    MVM_free(buffer);
                     /* Throw if it's unmapped */
                     MVM_exception_throw_adhoc(tc,
                         "Error decoding shiftjis string: could not decode byte 0x%hhX",
@@ -272,6 +284,7 @@ MVMString * MVM_string_shiftjis_decode(MVMThreadContext *tc,
                 continue;
             }
             else {
+                MVM_free(buffer);
                 MVM_exception_throw_adhoc(tc, "shiftjis decoder encountered an internal error.\n");
             }
         }
@@ -285,25 +298,30 @@ MVMString * MVM_string_shiftjis_decode(MVMThreadContext *tc,
             }
             last_was_cr = 0;
         }
-        else if (graph == '\r') {
+        /* Only follow this path if we aren't in a replacement
+         * Any replacement will already be NFG, and if we had a \r
+         * in the replacement, this will cause a loop */
+        else if (graph == '\r' && !start_repl) {
             last_was_cr = 1;
             continue;
         }
         if (result_graphs == result_size) {
             result_size += repl_length;
-            result->body.storage.blob_32 = MVM_realloc(result->body.storage.blob_32,
-                result_size * sizeof(MVMGrapheme32));
+            buffer = MVM_realloc(buffer, result_size * sizeof(MVMGrapheme32));
         }
-        result->body.storage.blob_32[result_graphs++] = graph;
+        buffer[result_graphs++] = graph;
     }
     /* If we end up with Shift_JIS_lead still set, that means we're missing a byte
      * that should have followed it. */
     if (Shift_JIS_lead != 0x00) {
+        MVM_free(buffer);
         MVM_exception_throw_adhoc(tc, "Error, ended decode of shiftjis expecting another byte. "
             "Last byte seen was 0x%hhX\n", Shift_JIS_lead);
     }
-    result->body.storage.blob_32 = MVM_realloc(result->body.storage.blob_32,
+    result = (MVMString *)REPR(result_type)->allocate(tc, STABLE(result_type));
+    result->body.storage.blob_32 = MVM_realloc(buffer,
         result_graphs * sizeof(MVMGrapheme32));
+    result->body.storage_type = MVM_STRING_GRAPHEME_32;
     result->body.num_graphs = result_graphs;
 
     return result;
@@ -322,6 +340,7 @@ MVMuint32 MVM_string_shiftjis_decodestream(MVMThreadContext *tc, MVMDecodeStream
     MVMuint32 reached_stopper;
     MVMStringIndex repl_length = ds->replacement ? MVM_string_graphs(tc, ds->replacement) : 0;
     MVMStringIndex repl_pos = 0;
+    int start_repl = 0;
     MVMuint8 Shift_JIS_lead = 0x00;
     MVMuint8 prepended = 0;
     int is_prepended = 0;
@@ -345,26 +364,40 @@ MVMuint32 MVM_string_shiftjis_decodestream(MVMThreadContext *tc, MVMDecodeStream
         /* Process this buffer. */
         MVMint32  pos = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
         MVMuint8 *bytes = (MVMuint8 *)cur_bytes->bytes;
-        while (pos < cur_bytes->length || repl_pos) {
+        while (pos < cur_bytes->length || start_repl) {
             MVMGrapheme32 graph;
-            MVMCodepoint codepoint;
+            MVMCodepoint codepoint = 0;
             MVMuint8 byte;
-            int graph_is_set = 0;
             int handler_rtrn = 0;
-            if (repl_pos) {
+            if (start_repl) {
+                /* Except for the case of the previous grapheme being \r and the replacement
+                 * starting with \n, we shouldn't get any normalization or codepoints combining
+                 * to make graphemes */
                 graph = MVM_string_get_grapheme_at_nocheck(tc, ds->replacement, repl_pos++);
-                graph_is_set = 1;
-                if (repl_length <= repl_pos) repl_pos = 0;
-            }
-            else if (is_prepended) {
-                byte = prepended;
-                is_prepended = 0;
+                if (last_was_cr) {
+                    if (graph == '\n') {
+                        graph = MVM_nfg_crlf_grapheme(tc);
+                    }
+                    else {
+                        graph = '\r';
+                        repl_pos--;
+                    }
+                    last_was_cr = 0;
+                }
+                if (repl_length <= repl_pos) {
+                    repl_pos   = 0;
+                    start_repl = 0;
+                }
             }
+            /* Main path. There was no replacement */
             else {
-                byte = bytes[pos++];
-            }
-            /* graph_is_set will be 0 unless we just grabbed a replacement grapheme */
-            if (!graph_is_set) {
+                if (is_prepended) {
+                    byte = prepended;
+                    is_prepended = 0;
+                }
+                else {
+                    byte = bytes[pos++];
+                }
                 handler_rtrn = decoder_handler(tc, &Shift_JIS_lead, byte, &codepoint);
                 if (handler_rtrn == DECODE_CODEPOINT) {
                     graph = codepoint;
@@ -377,11 +410,8 @@ MVMuint32 MVM_string_shiftjis_decodestream(MVMThreadContext *tc, MVMDecodeStream
                      * a replacement is used. */
                     Shift_JIS_lead = 0x00;
                     if (ds->replacement) {
-                        graph = MVM_string_get_grapheme_at_nocheck(tc, ds->replacement, repl_pos);
-                        /* If the replacement is more than one grapheme we need
-                         * to set repl_pos++ so we will grab the next grapheme on
-                         * the next loop */
-                        if (1 < repl_length) repl_pos++;
+                        start_repl = 1;
+                        continue;
                     }
                     else {
                         /* Throw if it's unmapped */
@@ -397,6 +427,7 @@ MVMuint32 MVM_string_shiftjis_decodestream(MVMThreadContext *tc, MVMDecodeStream
                     continue;
                 }
                 else {
+                    MVM_free(buffer);
                     MVM_exception_throw_adhoc(tc, "shiftjis decoder encountered an internal error. This bug should be reported.\n");
                 }
             }
@@ -410,7 +441,10 @@ MVMuint32 MVM_string_shiftjis_decodestream(MVMThreadContext *tc, MVMDecodeStream
                 }
                 last_was_cr = 0;
             }
-            else if (graph == '\r') {
+            /* Only follow this path if we aren't in a replacement
+             * Any replacement will already be NFG, and if we had a \r
+             * in the replacement, this will cause a loop */
+            else if (graph == '\r' && !start_repl) {
                 last_was_cr = 1;
                 continue;
             }
